# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Internationalization helpers."""

from oslo_config import cfg
import oslo_i18n._message as _msg

cfg.CONF.register_opts([
    cfg.BoolOpt('translation_lazy_load', default=True,
                help='Specifies whether to Lazy-Load Translation')
])


def _(msg, domain='pypowervm'):
    """Shortcut method to return a translated message."""
    msg = _msg.Message(msg, domain=domain)
    # If we are supposed to lazy-load, then return the message object
    return msg if cfg.CONF.translation_lazy_load else msg.translate()



# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import pypowervm.const as pc
import pypowervm.exceptions as ex
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.storage as stor
import pypowervm.wrappers.virtual_io_server as vios


class TestVolumeGroup(twrap.TestWrapper):

    file = 'fake_volume_group.txt'
    wrapper_class_to_test = stor.VG

    def test_base(self):
        """Tests baseline function within the Volume Group."""
        self.assertEqual('image_pool', self.dwrap.name)
        self.assertEqual(1063.3, self.dwrap.capacity)
        self.assertEqual(1051.1, self.dwrap.available_size)
        self.assertEqual(1051.2, self.dwrap.free_space)
        self.assertEqual('00f8d6de00004b000000014a54555cd9',
                         self.dwrap.serial_id)

    def test_vmedia_repos(self):
        """Tests the virtual media repositories."""
        repos = self.dwrap.vmedia_repos
        self.assertEqual(1, len(repos))
        self.assertEqual('VMLibrary', repos[0].name)
        self.assertEqual(11, repos[0].size)

        # Optical media
        vopts = repos[0].optical_media
        self.assertEqual(2, len(vopts))

        self.assertEqual('blank_media1', vopts[0].media_name)
        self.assertEqual('blank_media1', vopts[0].name)
        self.assertEqual(0.0977, vopts[0].size)
        self.assertEqual('0eblank_media1', vopts[0].udid)
        self.assertEqual('rw', vopts[0].mount_type)

    def test_physical_volumes(self):
        """Tests the physical volumes in the VG."""
        pvs = self.dwrap.phys_vols
        self.assertEqual(1, len(pvs))

        pv = pvs[0]
        self.assertEqual('01MUlCTSAgICAgSVBSLTAgICA1RDgyODMwMDAwMDAwMDQw',
                         pv.udid)
        self.assertEqual(1089592, pv.capacity)
        self.assertEqual('hdisk1', pv.name)
        self.assertEqual('active', pv.state)
        self.assertFalse(pv.is_fc_backed)
        self.assertTrue(pv.avail_for_use)
        self.assertEqual('SAS RAID 0 Disk Array', pv.description)
        self.assertEqual('U78C9.001.WZS0095-P1-C14-R1-L405D828300-L0',
                         pv.loc_code)

    def test_virtual_disk(self):
        """Tests the virtual disk gets."""
        vdisks = self.dwrap.virtual_disks
        self.assertEqual(1, len(vdisks))

        vdisk = vdisks[0]
        self.assertEqual('asdf', vdisk.name)
        self.assertEqual('None', vdisk.label)
        self.assertEqual(1, vdisk.capacity)
        self.assertEqual('0300f8d6de00004b000000014a54555cd9.1',
                         vdisk.udid)
        self.assertEqual(
            "https://9.1.2.3:12443/rest/api/uom/VirtualIOServer/14B854F7-42CE-"
            "4FF0-BD57-1D117054E701/VolumeGroup/b6bdbf1f-eddf-3c81-8801-9859eb"
            "6fedcb", vdisk.vg_uri)

        # Test setters
        vdisk.capacity = 2
        self.assertEqual(2, vdisk.capacity)
        vdisk.name = 'new_name'
        self.assertEqual('new_name', vdisk.name)
        vdisk._base_image('base_image')
        self.assertEqual('base_image', vdisk._get_val_str(stor._DISK_BASE))

    def test_add_vdisk(self):
        """Performs a test flow that adds a virtual disk."""
        vdisks = self.dwrap.virtual_disks

        self.assertEqual(1, len(vdisks))

        disk = stor.VDisk.bld(None, 'disk_name', 10.9876543, label='label',
                              base_image='cache')
        self.assertIsNotNone(disk)

        vdisks.append(disk)
        self.dwrap.virtual_disks = vdisks

        self.assertEqual(2, len(self.dwrap.virtual_disks))

        # make sure the second virt disk matches what we put in
        vdisk = self.dwrap.virtual_disks[1]
        self.assertEqual('disk_name', vdisk.name)
        self.assertEqual(10.987654, vdisk.capacity)
        self.assertEqual('label', vdisk.label)
        self.assertEqual(None, vdisk.udid)
        self.assertEqual('cache', vdisk._get_val_str(stor._DISK_BASE))

        # Try a remove
        self.dwrap.virtual_disks.remove(vdisk)
        self.assertEqual(1, len(self.dwrap.virtual_disks))

    def test_add_phys_vol(self):
        """Performs a test flow that adds a physical volume to the vol grp."""
        phys_vols = self.dwrap.phys_vols

        self.assertEqual(1, len(phys_vols))

        phys_vol = stor.PV.bld(None, 'disk1')
        self.assertIsNotNone(phys_vol)

        phys_vols.append(phys_vol)
        self.dwrap.phys_vols = phys_vols

        self.assertEqual(2, len(self.dwrap.phys_vols))

        # Make sure that the second physical volume matches
        pvol = self.dwrap.phys_vols[1]
        self.assertEqual('disk1', pvol.name)

    def test_add_media_repo(self):
        """Performs a simple add to the volume group of a new media repo."""
        media_repos = self.dwrap.vmedia_repos

        self.assertEqual(1, len(media_repos))

        vmedia_repo = stor.VMediaRepos.bld(None, 'repo', 10.12345)
        self.assertIsNotNone(vmedia_repo)

        media_repos.append(vmedia_repo)
        self.dwrap.vmedia_repos = media_repos

        self.assertEqual(2, len(self.dwrap.vmedia_repos))

        # Make sure that the second media repo matches
        repo = self.dwrap.vmedia_repos[1]
        self.assertEqual('repo', repo.name)
        self.assertEqual(10.12345, repo.size)
        self.assertEqual(0, len(repo.optical_media))

    def test_update_media_repo(self):
        """Performs a simple test to add optical media to an existing repo."""
        media_repos = self.dwrap.vmedia_repos

        vopt_medias = media_repos[0].optical_media
        self.assertEqual(2, len(vopt_medias))

        new_media = stor.VOptMedia.bld(None, 'name', 0.123, 'r')
        self.assertIsNotNone(new_media)

        vopt_medias.append(new_media)
        media_repos[0].optical_media = vopt_medias
        self.assertEqual(3, len(media_repos[0].optical_media))

        # Check the attributes
        media = media_repos[0].optical_media[2]
        self.assertEqual('name', media.media_name)
        self.assertEqual('name', media.name)
        self.assertEqual(0.123, media.size)
        self.assertEqual('r', media.mount_type)

    def test_ordering(self):
        """Set fields out of order; ensure they end up in the right order."""
        vg = stor.VG._bld(None)
        vg.virtual_disks = []
        vg.name = 'vgname'
        vg.vmedia_repos = []
        vg.set_parm_value(stor._VG_CAPACITY, 123)
        vg.phys_vols = []
        self.assertEqual(
            vg.toxmlstring(),
            '<uom:VolumeGroup xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:GroupCapacity>123</uom:GroupCapacity'
            '><uom:GroupName>vgname</uom:GroupName><uom:MediaRepositories sche'
            'maVersion="V1_0"/><uom:PhysicalVolumes schemaVersion="V1_0"/><uom'
            ':VirtualDisks schemaVersion="V1_0"/></uom:VolumeGroup>'.
            encode('utf-8'))

    def test_bld(self):
        vg = stor.VG.bld(None, 'myvg', [stor.PV.bld(None, 'hdisk1')])
        self.assertEqual(
            vg.toxmlstring(),
            '<uom:VolumeGroup xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:GroupName>myvg</uom:GroupName><uom:P'
            'hysicalVolumes schemaVersion="V1_0"><uom:PhysicalVolume schemaVer'
            'sion="V1_0"><uom:Metadata><uom:Atom/></uom:Metadata><uom:VolumeNa'
            'me>hdisk1</uom:VolumeName></uom:PhysicalVolume></uom:PhysicalVolu'
            'mes></uom:VolumeGroup>'.encode('utf-8'))


class TestLUEnt(twrap.TestWrapper):
    file = 'lufeed.txt'
    wrapper_class_to_test = stor.LUEnt

    def test_logical_units(self):
        lus = self.entries
        self.assertEqual(len(lus), 22)
        lu = lus[0]
        self.assertEqual('a2b11d20-322d-3dcc-84ee-74fce7e90310', lu.uuid)
        self.assertEqual(lu.udid, '276c097502d44311e58004000040f2e95dc4a789dfd'
                                  '63464654319f89c04aa5382')
        self.assertEqual(lu.name, 'boot_pvm3_tempest_ser_a73d7083')
        self.assertTrue(lu.is_thin)
        self.assertEqual(lu.lu_type, 'VirtualIO_Disk')
        self.assertAlmostEqual(lu.capacity, 1, 1)
        self.assertFalse(lu.in_use)
        self.assertEqual('296c097502d44311e58004000040f2e95dcf6063cd6bf7ed7b49'
                         '7780a1799236ab', lu.cloned_from_udid)

    def test_lu_bld(self):
        lu = stor.LUEnt.bld(None, 'lu_name', 123)
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:UnitCapacity>123.000000</uom:UnitCap'
            'acity><uom:UnitName>lu_name</uom:UnitName></uom:LogicalUnit>'.
            encode('utf-8'))
        lu = stor.LUEnt.bld(None, 'lu_name', 1.2345678, thin=True,
                            typ=stor.LUType.IMAGE)
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:ThinDevice>true</uom:ThinDevice><uom'
            ':UnitCapacity>1.234568</uom:UnitCapacity><uom:LogicalUnitType>Vir'
            'tualIO_Image</uom:LogicalUnitType><uom:UnitName>lu_name</uom:Unit'
            'Name></uom:LogicalUnit>'.encode('utf-8'))
        lu = stor.LUEnt.bld(None, 'lu_name', .12300019999, thin=False,
                            clone=mock.Mock(capacity=1.23,
                                            udid='cloned_from_udid'))
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:ThinDevice>false</uom:ThinDevice><uo'
            'm:UnitCapacity>1.230000</uom:UnitCapacity><uom:ClonedFrom>cloned_'
            'from_udid</uom:ClonedFrom><uom:UnitName>lu_name</uom:UnitName></u'
            'om:LogicalUnit>'.encode('utf-8'))

    def test_lu_ordering(self):
        lu = stor.LUEnt._bld(None)
        lu._name('lu_name')
        lu._udid('lu_udid')
        lu._cloned_from_udid('cloned_from')
        lu._capacity(123)
        lu.set_parm_value(stor._LU_THIN, 'true')
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:ThinDevice>true</uom:ThinDevice><uom'
            ':UniqueDeviceID>lu_udid</uom:UniqueDeviceID><uom:UnitCapacity>123'
            '.000000</uom:UnitCapacity><uom:ClonedFrom>cloned_from</uom:Cloned'
            'From><uom:UnitName>lu_name</uom:UnitName></uom:LogicalUnit>'.
            encode('utf-8'))

    def test_lu_equality(self):
        # We can equate LUEnt and LU that represent the same LogicalUnit
        lu1 = stor.LUEnt.bld(None, 'mylu', 1)
        lu2 = stor.LU.bld(None, 'mylu', 2)
        self.assertEqual(lu1, lu2)
        lu1._udid('lu_udid')
        lu2._udid('lu_udid')
        self.assertEqual(lu1, lu2)
        lu2._udid('another_udid')
        self.assertNotEqual(lu1, lu2)
        lu2._udid('lu_udid')
        lu1._name('another_lu')
        self.assertNotEqual(lu1, lu2)

    def test_lu_hash(self):
        udid1 = ('27cfc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d8')
        # Only prefix differs.  Should fail == but hash equal
        udid2 = ('29cfc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d8')
        # Last bit differs
        udid3 = ('27cfc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d9')
        # First bit differs
        udid4 = ('274fc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d8')
        lu1 = stor.LUEnt.bld(None, 'mylu', 1)
        lu2 = stor.LUEnt.bld(None, 'mylu', 2)
        lu1._udid(udid1)
        lu2._udid(udid1)
        self.assertEqual({lu1}, {lu2})
        lu2._udid(udid2)
        self.assertNotEqual({lu1}, {lu2})
        self.assertEqual(hash(lu1), hash(lu2))
        lu2._udid(udid3)
        self.assertNotEqual({lu1}, {lu2})
        lu2._udid(udid4)
        self.assertNotEqual({lu1}, {lu2})


class TestTier(twrap.TestWrapper):
    file = 'tier.txt'
    wrapper_class_to_test = stor.Tier

    def test_props(self):
        self.assertEqual(1, len(self.entries))
        tier = self.dwrap
        self.assertEqual('SYSTEM', tier.name)
        self.assertEqual('256c097502d44311e58004000040f2e95d7d95846d854f9f38',
                         tier.udid)
        self.assertTrue(tier.is_default)
        self.assertAlmostEqual(3071.25, tier.capacity)
        self.assertEqual('535e1e51-50a6-3722-b6ed-907ff011a535', tier.ssp_uuid)


class TestSharedStoragePool(twrap.TestWrapper):

    file = 'ssp.txt'
    wrapper_class_to_test = stor.SSP

    def test_name(self):
        self.assertEqual(self.dwrap.name, 'neossp1')

    def test_udid(self):
        self.assertEqual(self.dwrap.udid, '24cfc907d2abf511e4b2d540f2e95daf3'
                         '0000000000972FB370000000054D14EB8')

    def test_capacity(self):
        self.assertAlmostEqual(self.dwrap.capacity, 49.88, 3)

    def test_free_space(self):
        self.assertAlmostEqual(self.dwrap.free_space, 48.98, 3)

    def test_over_commit_space(self):
        self.assertEqual(self.dwrap.over_commit_space, 1.234567)

    def test_total_lu_size(self):
        self.assertAlmostEqual(self.dwrap.total_lu_size, 1, 1)

    def test_physical_volumes(self):
        pvs = self.dwrap.physical_volumes
        self.assertEqual(len(pvs), 1)
        pv = pvs[0]
        self.assertEqual(
            pv.udid,
            '01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEODgwMDAwMDAwMDAwMDAwMw==')
        self.assertEqual(pv.name, 'hdisk3')
        # Test setter
        self.dwrap.physical_volumes = []
        pvs = self.dwrap.physical_volumes
        self.assertEqual(0, len(pvs))
        self.dwrap.physical_volumes = [stor.PV.bld(None, 'pv1', udid='udid1'),
                                       stor.PV.bld(None, 'pv2', udid='udid2')]
        pvs = self.dwrap.physical_volumes
        self.assertEqual(2, len(pvs))
        self.assertEqual(dict(pv1='udid1', pv2='udid2'),
                         {pv.name: pv.udid for pv in pvs})

    def test_logical_units(self):
        lus = self.dwrap.logical_units
        self.assertEqual(len(lus), 1)
        lu = lus[0]
        self.assertEqual(lu.udid, '27cfc907d2abf511e4b2d540f2e95daf301a02b090'
                         '4778d755df5a46fe25e500d8')
        self.assertEqual(lu.name, 'neolu1')
        self.assertTrue(lu.is_thin)
        self.assertEqual(lu.lu_type, 'VirtualIO_Disk')
        self.assertAlmostEqual(lu.capacity, 1, 1)
        self.assertTrue(lu.in_use)
        # Test setter
        self.dwrap.logical_units = []
        lus = self.dwrap.logical_units
        self.assertEqual(0, len(lus))
        self.dwrap.logical_units = [stor.LU.bld(None, 'lu1', 1),
                                    stor.LU.bld(None, 'lu2', 2)]
        lus = self.dwrap.logical_units
        self.assertEqual(2, len(lus))
        self.assertEqual(dict(lu1=1, lu2=2),
                         {lu.name: lu.capacity for lu in lus})

    def test_fresh_ssp(self):
        ssp = stor.SSP.bld(None, 'myssp', [
            stor.PV.bld(None, name=n) for n in (
                'hdisk123', 'hdisk132', 'hdisk213', 'hdisk231', 'hdisk312',
                'hdisk321')])
        self.assertEqual(ssp.name, 'myssp')
        self.assertEqual(ssp.schema_type, stor.SSP.schema_type)
        self.assertEqual(ssp.schema_ns, pc.UOM_NS)
        pvs = ssp.physical_volumes
        self.assertEqual(len(pvs), 6)
        pv = pvs[3]  # hdisk231
        self.assertEqual(pv.schema_type, stor.PV.schema_type)
        self.assertEqual(pv.schema_ns, pc.UOM_NS)
        self.assertEqual(pv.name, 'hdisk231')

    def test_lu_bld(self):
        lu = stor.LU.bld(None, 'lu_name', 123)
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:UnitCapacity>123.000000</uom:UnitCap'
            'acity><uom:UnitName>lu_name</uom:UnitName></uom:LogicalUnit>'.
            encode('utf-8'))
        lu = stor.LU.bld(None, 'lu_name', 1.2345678, thin=True)
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:ThinDevice>true</uom:ThinDevice><uom'
            ':UnitCapacity>1.234568</uom:UnitCapacity><uom:UnitName>lu_name</u'
            'om:UnitName></uom:LogicalUnit>'.encode('utf-8'))
        lu = stor.LU.bld(None, 'lu_name', .12300019999, thin=False)
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:ThinDevice>false</uom:ThinDevice><uo'
            'm:UnitCapacity>0.123000</uom:UnitCapacity><uom:UnitName>lu_name</'
            'uom:UnitName></uom:LogicalUnit>'.encode('utf-8'))

    def test_lu_ordering(self):
        lu = stor.LU._bld(None)
        lu._name('lu_name')
        lu._udid('lu_udid')
        lu.set_parm_value(stor._LU_CLONED_FROM, 'cloned_from')
        lu._capacity(123)
        lu.set_parm_value(stor._LU_THIN, 'true')
        self.assertEqual(
            lu.toxmlstring(),
            '<uom:LogicalUnit xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
            'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata><u'
            'om:Atom/></uom:Metadata><uom:ThinDevice>true</uom:ThinDevice><uom'
            ':UniqueDeviceID>lu_udid</uom:UniqueDeviceID><uom:UnitCapacity>123'
            '.000000</uom:UnitCapacity><uom:ClonedFrom>cloned_from</uom:Cloned'
            'From><uom:UnitName>lu_name</uom:UnitName></uom:LogicalUnit>'.
            encode('utf-8'))

    def test_lu_equality(self):
        lu1 = stor.LU.bld(None, 'mylu', 1)
        lu2 = stor.LU.bld(None, 'mylu', 2)
        self.assertEqual(lu1, lu2)
        lu1._udid('lu_udid')
        lu2._udid('lu_udid')
        self.assertEqual(lu1, lu2)
        lu2._udid('another_udid')
        self.assertNotEqual(lu1, lu2)
        lu2._udid('lu_udid')
        lu1._name('another_lu')
        self.assertNotEqual(lu1, lu2)

    def test_lu_hash(self):
        udid1 = ('27cfc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d8')
        # Only prefix differs.  Should fail == but hash equal
        udid2 = ('29cfc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d8')
        # Last bit differs
        udid3 = ('27cfc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d9')
        # First bit differs
        udid4 = ('274fc907d2abf511e4b2d540f2e95daf3'
                 '01a02b0904778d755df5a46fe25e500d8')
        lu1 = stor.LU.bld(None, 'mylu', 1)
        lu2 = stor.LU.bld(None, 'mylu', 2)
        lu1._udid(udid1)
        lu2._udid(udid1)
        self.assertEqual({lu1}, {lu2})
        lu2._udid(udid2)
        self.assertNotEqual({lu1}, {lu2})
        self.assertEqual(hash(lu1), hash(lu2))
        lu2._udid(udid3)
        self.assertNotEqual({lu1}, {lu2})
        lu2._udid(udid4)
        self.assertNotEqual({lu1}, {lu2})


class TestVFCClientAdapter(twrap.TestWrapper):
    file = 'vfc_client_adapter_feed.txt'
    wrapper_class_to_test = stor.VFCClientAdapter

    def test_vfc_client_adapter(self):
        """Check getters on VFCClientAdapter.

        The hard part - the wrapping - was done by TestWrapper.
        """
        self.assertEqual('U8247.21L.212A64A-V25-C4', self.dwrap.loc_code)
        self.assertEqual(25, self.dwrap.lpar_id)
        self.assertEqual(2, self.dwrap.vios_id)
        self.assertEqual('Client', self.dwrap.side)
        self.assertEqual(4, self.dwrap.lpar_slot_num)
        self.assertEqual(10, self.dwrap.vios_slot_num)
        self.assertEqual(['C05076087CBA0169', 'C05076087CBA0168'],
                         self.dwrap.wwpns)


class TestVIOS(twrap.TestWrapper):
    file = 'vio_multi_vscsi_mapping.txt'
    wrapper_class_to_test = vios.VIOS

    def test_pg83_in_pv(self):
        """Legitimate pg83 data from the <PhysicalVolume/>."""
        self.assertEqual('600507680282861D88000000000000B5',
                         self.dwrap.phys_vols[1].pg83)

    # TODO(efried): reinstate when VIOS supports pg83 descriptor in Events
    # def test_pg83_absent_from_pv(self):
    #     """No pg83 data in <PhysicalVolume/>."""
    #     self.assertIsNone(self.dwrap.phys_vols[0].pg83)

    @mock.patch('pypowervm.wrappers.job.Job.wrap')
    def test_pg83_absent_from_pv(self, mock_wrap):
        """LUARecovery.QUERY_INVENTORY when no pg83 in <PhysicalVolume/>."""
        # TODO(efried): remove this method once VIOS supports pg83 in Events
        mock_jwrap = mock.Mock()
        mock_jwrap.get_job_results_as_dict.return_value = {
            'OutputXML':
                '<VIO version="1.21" xmlns="http://ausgsa.austin.ibm">'
                '<Response><InventoryResponse viosId="8247-21L03212A60A" '
                'sequence="435" inventoryType="base" eventLogOn="true">'
                '<PhysicalVolume udid="01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEO'
                'DgwMDAwMDAwMDAwMDBEQQ==" name="hdisk2" '
                'description="MPIO IBM 2076 FC Disk">'
                '<PhysicalVolume_base capacity="20480" '
                'locationCode="U78CB.001.WZS05HN-P1-C7-T1-W500507680220E523-L2'
                '000000000000" '
                'unique_id="33213600507680282861D880000000DA04214503IBMfcp" '
                'descriptor="SV9hbV9hX3BnODNfTkFBX2Rlc2NyaXB0b3I=" '
                'desType="NAA"></PhysicalVolume_base></PhysicalVolume>'
                '</InventoryResponse></Response></VIO><?xml version="1.0"?>'
                '<uom:VIO xmlns:uom="http://www.ibm.com/xmlns/systems/power/fi'
                'rmware/uom/mc/2012_10/" xmlns="" version="1.21">'
                '<uom:Response/></uom:VIO>'}
        mock_wrap.return_value = mock_jwrap
        self.assertEqual('I_am_a_pg83_NAA_descriptor',
                         self.dwrap.phys_vols[0].pg83)
        mock_jwrap.run_job.assert_called_with(
            '3443DB77-AED1-47ED-9AA5-3DB9C6CF7089', job_parms=[mock.ANY])

    def test_pg83_raises_if_no_parent_entry(self):
        """Raise attempting to get pg83 if PV has no parent_entry."""
        # TODO(efried): remove this method once VIOS supports pg83 in Events
        pv = stor.PV.bld(self.adpt, 'name', 'udid')
        self.assertRaises(ex.UnableToBuildPG83EncodingMissingParent,
                          lambda: pv.pg83)

    def test_bogus_pg83_in_pv(self):
        """Bogus pg83 data in the <PhysicalVolume/> doesn't trigger the Job."""
        with self.assertLogs(stor.__name__, 'WARNING'):
            self.assertIsNone(self.dwrap.phys_vols[2].pg83)


class TestTargetDevs(twrap.TestWrapper):
    """Tests for VSCSIMapping.target_dev and {storage_type}TargetDev wrappers.

    SCSI mapping target devices in the test file are laid out as follows:

    index  type            LUA
    0      LUTargetDev     0x8200000000000000
    1      None            -
    2      None            -
    3      LUTargetDev     0x8400000000000000
    4      LUTargetDev     0x8500000000000000
    5      VOptTargetDev   0x8100000000000000
    6      LUTargetDev     0x8300000000000000
    7      PVTargetDev     0x8600000000000000
    8      VDiskTargetDev  0x8700000000000000
    """
    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def test_subtypes_props(self):
        """Right subtypes and LUA gets/sets from VSCSIMapping.target_dev."""
        smaps = self.dwrap.scsi_mappings
        # LU
        self.assertIsInstance(smaps[0].target_dev, stor.LUTargetDev)
        self.assertEqual('0x8200000000000000', smaps[0].target_dev.lua)
        smaps[0].target_dev._lua('lu_lua')
        self.assertEqual('lu_lua', smaps[0].target_dev.lua)
        # No TargetDevice
        self.assertIsNone(smaps[1].target_dev)
        # VOpt
        self.assertIsInstance(smaps[5].target_dev, stor.VOptTargetDev)
        self.assertEqual('0x8100000000000000', smaps[5].target_dev.lua)
        smaps[5].target_dev._lua('vopt_lua')
        self.assertEqual('vopt_lua', smaps[5].target_dev.lua)
        # PV
        self.assertIsInstance(smaps[7].target_dev, stor.PVTargetDev)
        self.assertEqual('0x8600000000000000', smaps[7].target_dev.lua)
        smaps[7].target_dev._lua('pv_lua')
        self.assertEqual('pv_lua', smaps[7].target_dev.lua)
        # LV/VDisk
        self.assertIsInstance(smaps[8].target_dev, stor.VDiskTargetDev)
        self.assertEqual('0x8700000000000000', smaps[8].target_dev.lua)
        smaps[8].target_dev._lua('lv_lua')
        self.assertEqual('lv_lua', smaps[8].target_dev.lua)

    def test_bld_set(self):
        """Test *TargetDev.bld(lua) and setting VSCSIMapping.target_dev."""
        smaps = self.dwrap.scsi_mappings
        for klass in (stor.LUTargetDev, stor.VOptTargetDev, stor.PVTargetDev,
                      stor.VDiskTargetDev):
            lua_tag = klass.__class__.__name__ + "_lua"
            # Build LU target dev
            vtd = klass.bld('adap', lua_tag)
            self.assertEqual('adap', vtd.adapter)
            self.assertEqual(lua_tag, vtd.lua)
            # Assign to a target_dev in the SCSI mappings.  Pick one that's
            # initially empty: the first iteration will prove we can assign to
            # an empty one; subsequent iterations will prove we can overwrite.
            smaps[1]._target_dev(vtd)
            self.assertIsInstance(smaps[1].target_dev, klass)
            self.assertEqual(lua_tag, smaps[1].target_dev.lua)


class TestStorageTypes(testtools.TestCase):
    def test_fileio(self):
        fio = stor.FileIO.bld_ref('adap', 'path')
        self.assertEqual('path', fio.label)
        self.assertEqual('path', fio.path)
        self.assertEqual('path', fio.name)
        self.assertIsNone(fio.capacity)
        self.assertIsNone(fio.udid)
        self.assertEqual('File', fio.vdtype)
        self.assertIsNone(fio.backstore_type)
        # Explicit backstore, legacy bld method
        fio = stor.FileIO.bld('adap', 'path',
                              backstore_type=stor.BackStoreType.FILE_IO)
        self.assertEqual('path', fio.label)
        self.assertEqual('path', fio.path)
        self.assertEqual('path', fio.name)
        self.assertIsNone(fio.capacity)
        self.assertIsNone(fio.udid)
        self.assertEqual('File', fio.vdtype)
        self.assertEqual('fileio', fio.backstore_type)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks around IBMi VM changes."""

from oslo_log import log as logging

import pypowervm.const as c
import pypowervm.exceptions as pvmex
from pypowervm import i18n
import pypowervm.tasks.scsi_mapper as pvm_smap
import pypowervm.tasks.vfc_mapper as pvm_vfcmap
import pypowervm.wrappers.base_partition as pvm_bp
import pypowervm.wrappers.logical_partition as pvm_lpar
from pypowervm.wrappers import virtual_io_server as pvm_vios

LOG = logging.getLogger(__name__)

# TODO(IBM) translation
_LI = i18n._


def update_ibmi_settings(adapter, lpar_w, boot_type):
    """Update TaggedIO, Keylock postion and IPL Source of IBMi VM.

    TaggedIO of IBMi vm will be updated to identify the load source,
    alternative load source and console type. Keylock position will be set
    to the value of NORMAL in KeylockPos enumration. IPL Source will be set
    to the value of B in IPLSrc enumration.
    :param adapter: The pypowervm adapter.
    :param lpar_w: The lpar wrapper.
    :param boot_type: The boot connectivity type of the VM. It is a string
                      value that represents one of the values in the
                      BootStorageType enumeration.
    :return: The updated LPAR wrapper. The update is not executed against the
             system, but rather the wrapper itself is updated locally.
    """
    load_source = None
    alt_load_source = None
    client_adapters = []
    if boot_type == pvm_lpar.BootStorageType.VFC:
        msg = _LI("Setting Virtual Fibre Channel slot as load source for VM "
                  "%s") % lpar_w.name
        LOG.info(msg)
        for vios_wrap in pvm_vios.VIOS.get(adapter, xag=[c.XAG.VIO_FMAP]):
            existing_maps = pvm_vfcmap.find_maps(
                vios_wrap.vfc_mappings, lpar_w.id)
            client_adapters.extend([vfcmap.client_adapter
                                    for vfcmap in existing_maps
                                    if vfcmap.client_adapter is not None])
    else:
        # That boot volume, which is vscsi physical volume, ssp lu
        # and local disk, could be handled here.
        msg = _LI("Setting Virtual SCSI slot slot as load source for VM "
                  "%s") % lpar_w.name
        LOG.info(msg)
        for vios_wrap in pvm_vios.VIOS.get(adapter, xag=[c.XAG.VIO_SMAP]):
            existing_maps = pvm_smap.find_maps(
                vios_wrap.scsi_mappings, lpar_w.id)
            client_adapters.extend([smap.client_adapter
                                    for smap in existing_maps
                                    if smap.client_adapter is not None])
    slot_nums = set(s.lpar_slot_num for s in client_adapters)
    slot_nums = list(slot_nums)
    slot_nums.sort()
    if len(slot_nums) > 0:
        load_source = slot_nums.pop(0)
    if len(slot_nums) > 0:
        alt_load_source = slot_nums.pop(0)
    if load_source is not None:
        if alt_load_source is None:
            alt_load_source = load_source
        lpar_w.io_config.tagged_io = pvm_bp.TaggedIO.bld(
            adapter, load_src=load_source, console='HMC',
            alt_load_src=alt_load_source)
    else:
        raise pvmex.IBMiLoadSourceNotFound(vm_name=lpar_w.name)
    lpar_w.desig_ipl_src = pvm_lpar.IPLSrc.B
    lpar_w.keylock_pos = pvm_bp.KeylockPos.NORMAL
    return lpar_w



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""General utilities around wrappers."""

import importlib
import os
import sys

import pypowervm.wrappers.entry_wrapper as ewrap

_wrappers_pkg = ('pypowervm', 'wrappers')

_imports = None

_this_module = sys.modules[__name__]


def _get_imports():
    """Imports all modules in pypowervm.wrappers and returns them as a dict.

    The returned dict is of the form:

        { module_name: <module object>, ... }
    """
    global _imports
    if _imports is None:
        _modnames = [fname.rsplit('.', 1)[0]
                     for fname in os.listdir(os.path.join(*_wrappers_pkg))
                     if not fname.startswith('_')
                     and fname.endswith('.py')]
        _imports = {
            modname: importlib.import_module(
                '.'.join(_wrappers_pkg) + '.' + modname)
            for modname in _modnames}
    return _imports


def wrapper_class_iter():
    """Iterator over all Wrapper subclasses defined in pypowervm.wrappers.

    Each yield is the Wrapper subclass itself.
    """
    for klass in (cls for imp in _get_imports().values()
                  for cls in vars(imp).values()):
        try:
            if issubclass(klass, ewrap.Wrapper):
                yield klass
        except TypeError:
            # issubclass can't handle the things that aren't classes
            pass



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Wrappers to parse the PCM JSON data from IBM.Host Resource Manager. """

import json


class LparInfo(object):
    """Represents a monitor sample from the IBM.Host Resource Manager.

    The Lpar JSON utilization data has the following structure:
    - LparUtil
      - LparMemory
    lpar_metrics are generally collected once every two minutes, as opposed
    to the other data which is collected every 30 seconds.
    """

    def __init__(self, raw_json):
        data = json.loads(raw_json)
        self._lparuuid_to_util_dict = dict()
        lpar_util_list = list()
        lpar_metric_list_rsct = data.get('lparUtil')
        for lpar_metrics_rsct in lpar_metric_list_rsct:
            if 'errorInfo' in lpar_metrics_rsct:
                error_code = lpar_metrics_rsct['errorInfo']['errorId']
                if error_code in ('6001', '6003'):
                    self._create_lpar_memory_util_for_errored_vm(
                        error_code, lpar_metrics_rsct)
                else:
                    # Any other errors that might get introduced at neo-rest.
                    continue
            lpar_util = LparUtil(lpar_metrics_rsct)
            lpar_util_list.append(lpar_util)
            self._lparuuid_to_util_dict[lpar_util.uuid] = lpar_util
        self._lpars_util = lpar_util_list

    def find(self, lpar_uuid):
        return self._lparuuid_to_util_dict.get(lpar_uuid, None)

    def _create_lpar_memory_util_for_errored_vm(
            self, error_code, lpar_metrics_rsct):
        if error_code == '6001':
            # If LPAR is powered off, then no memory is being used.
            memory = dict(pctRealMemFree=100,
                          vmPgInRate=0,
                          vmPgOutRate=0,
                          vmPgSpInRate=0,
                          vmPgSpOutRate=0)
        elif error_code == '6003':
            # If LPAR has inactive RMC, then assume all memory is being used.
            memory = dict(pctRealMemFree=0)
        lpar_metrics_rsct['memory'] = memory

    @property
    def lpars_util(self):
        return self._lpars_util


class LparUtil(object):
    """Represents individual Lpar metric information. """
    def __init__(self, lpar_util):
        self._uuid = lpar_util.get('uuid')
        self._lpar_id = lpar_util.get('id')
        self._name = lpar_util.get('name')
        self._timestamp = lpar_util.get('timestamp')
        self._memory = LparMemory(lpar_util.get('memory'))

    @property
    def lpar_id(self):
        return self._lpar_id

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def memory(self):
        return self._memory


class LparMemory(object):
    """Represents information on Lpar memory utilization """
    def __init__(self, memory):
        self._pct_real_mem_avbl = memory.get('pctRealMemAvbl')
        self._total_pg_count = memory.get('totalPgSpSizeCount')
        self._free_pg_count = memory.get('totalPgSpFreeCount')
        self._active_pg_count = memory.get('vmActivePgCount')
        self._real_mem_size_bytes = memory.get('realMemSizeBytes')
        self._pct_real_mem_free = memory.get('pctRealMemFree')
        self._vm_pg_in_rate = memory.get('vmPgInRate')
        self._vm_pg_out_rate = memory.get('vmPgOutRate')
        self._vm_pg_swap_in_rate = memory.get('vmPgSpInRate')
        self._vm_pg_swap_out_rate = memory.get('vmPgSpOutRate')

    @property
    def pct_real_mem_avbl(self):
        return self._pct_real_mem_avbl

    @property
    def total_pg_count(self):
        return self._total_pg_count

    @property
    def free_pg_count(self):
        return self._free_pg_count

    @property
    def active_pg_count(self):
        return self._active_pg_count

    @property
    def real_mem_size_bytes(self):
        return self._real_mem_size_bytes

    @property
    def pct_real_mem_free(self):
        return self._pct_real_mem_free

    @property
    def vm_pg_in_rate(self):
        return self._vm_pg_in_rate

    @property
    def vm_pg_out_rate(self):
        return self._vm_pg_out_rate

    @property
    def vm_pg_swap_in_rate(self):
        return self._vm_pg_swap_in_rate

    @property
    def vm_pg_swap_out_rate(self):
        return self._vm_pg_swap_out_rate



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import pypowervm.entities as ent
from pypowervm.tasks import master_mode as m_mode
import pypowervm.tests.tasks.util as u
import pypowervm.tests.test_fixtures as fx


class TestMasterMode(testtools.TestCase):
    """Unit Tests for master mode request and release."""

    def setUp(self):
        super(TestMasterMode, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        mock_resp = mock.MagicMock()
        mock_resp.entry = ent.Entry(
            {}, ent.Element('Dummy', self.adpt), self.adpt)
        self.adpt.read.return_value = mock_resp
        self.msys_w = mock.MagicMock()
        self.msys_w.adapter = self.adpt
        self.msys_w.uuid = '1234'

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    def test_request_master(self, mock_run_job):
        mock_run_job.side_effect = u.get_parm_checker(
            self, '1234', [(m_mode.CO_MGMT_MASTER_STATUS,
                            m_mode.MasterMode.NORMAL)], exp_timeout=1800)
        m_mode.request_master(self.msys_w)
        self.adpt.read.assert_called_once_with('ManagedSystem', '1234',
                                               suffix_parm='RequestMaster',
                                               suffix_type='do')
        self.adpt.reset_mock()
        mock_run_job.reset_mock()

        # Test temp mode
        mock_run_job.side_effect = u.get_parm_checker(
            self, '1234', [(m_mode.CO_MGMT_MASTER_STATUS,
                            m_mode.MasterMode.TEMP)], exp_timeout=1800)
        m_mode.request_master(self.msys_w, mode=m_mode.MasterMode.TEMP)
        self.adpt.read.assert_called_once_with('ManagedSystem', '1234',
                                               suffix_parm='RequestMaster',
                                               suffix_type='do')

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    def test_release_master(self, mock_run_job):
        m_mode.release_master(self.msys_w)
        self.adpt.read.assert_called_once_with('ManagedSystem', '1234',
                                               suffix_parm='ReleaseMaster',
                                               suffix_type='do')
        mock_run_job.assert_called_once_with('1234', timeout=1800)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""EntryWrappers for Cluster and its subelements."""

from oslo_log import log as logging

import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap
import pypowervm.wrappers.mtms as mtmwrap
import pypowervm.wrappers.storage as stor

LOG = logging.getLogger(__name__)

# Cluster Constants
_CL_NAME = 'ClusterName'
_CL_ID = 'ClusterID'
_CL_REPOPVS = 'RepositoryDisk'  # Yes, really
_CL_PV = stor.PHYS_VOL
_CL_SSP_LINK = 'ClusterSharedStoragePool'
_CL_NODES = 'Node'  # Yes, really
_CL_NODE = 'Node'
_CL_CAPABILITY = 'ClusterCapabilities'
_CL_EL_ORDER = (_CL_NAME, _CL_ID, _CL_REPOPVS, _CL_SSP_LINK, _CL_NODE,
                _CL_CAPABILITY)
# Node Constants
_N_HOSTNAME = 'HostName'
_N_LPARID = 'PartitionID'
_N_NAME = 'PartitionName'
_N_VIOS_LEVEL = 'VirtualIOServerLevel'
_N_VIOS_LINK = 'VirtualIOServer'
_N_IPADDR = 'IPAddress'
_N_STATE = 'State'
_N_EL_ORDER = (_N_HOSTNAME, _N_LPARID, _N_NAME, mtmwrap.MTMS_ROOT,
               _N_VIOS_LEVEL, _N_VIOS_LINK, _N_IPADDR, _N_STATE)


class NodeState(object):
    """Cluster node state, from NodeState.Enum."""
    UP = 'Up'
    DOWN = 'Down'
    UNKNOWN = 'Unknown'


@ewrap.EntryWrapper.pvm_type('Cluster', child_order=_CL_EL_ORDER)
class Cluster(ewrap.EntryWrapper):
    """A Cluster behind a SharedStoragePool."""

    search_keys = dict(name='ClusterName')

    @classmethod
    def bld(cls, adapter, name, repos_pv, first_node):
        """Create a fresh Cluster EntryWrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: String name for the Cluster.
        :param repos_pv: storage.PV representing the repository disk.
        :param first_node: Node wrapper representing the first VIOS
        to host the Cluster.  (The Cluster Create API only accepts a single
        node; others must be added later.)  The VIOS must be able to see each
        disk.
        """
        clust = cls._bld(adapter)
        clust.repos_pv = repos_pv
        clust.nodes = [first_node]
        clust._name(name)
        return clust

    @property
    def name(self):
        return self._get_val_str(_CL_NAME)

    def _name(self, newname):
        self.set_parm_value(_CL_NAME, newname)

    @property
    def id(self):
        """The string ID according to VIOS, not a UUID or UDID."""
        return self._get_val_str(_CL_ID)

    @property
    def ssp_uri(self):
        """The URI of the SharedStoragePool associated with this Cluster."""
        return self.get_href(_CL_SSP_LINK, one_result=True)

    @property
    def ssp_uuid(self):
        """The UUID of the SharedStoragePool associated with this Cluster."""
        uri = self.ssp_uri
        if uri is not None:
            return u.get_req_path_uuid(uri)

    @property
    def repos_pv(self):
        """Returns the (one) repository PV.

        Although the schema technically allows a collection of PVs under the
        RepositoryDisk element, a Cluster always has exactly one repository PV.
        """
        repos_elem = self._find_or_seed(_CL_REPOPVS)
        pv_list = repos_elem.findall(_CL_PV)
        # Check only relevant when building up a Cluster wrapper internally
        if pv_list and len(pv_list) == 1:
            return stor.PV.wrap(pv_list[0])
        return None

    @repos_pv.setter
    def repos_pv(self, pv):
        """Set the (single) PV member of RepositoryDisk.

        You cannot change the repository disk of a live Cluster.  This setter
        is useful only when constructing new Clusters.

        :param pv: The PV (NOT a list) to set.
        """
        self.replace_list(_CL_REPOPVS, [pv])

    @property
    def nodes(self):
        """WrapperElemList of Node wrappers."""
        return ewrap.WrapperElemList(self._find_or_seed(_CL_NODES), Node)

    @nodes.setter
    def nodes(self, ns):
        self.replace_list(_CL_NODES, ns)


@ewrap.ElementWrapper.pvm_type('Node', has_metadata=True,
                               child_order=_N_EL_ORDER)
class Node(ewrap.ElementWrapper):
    """A Node represents a VIOS member of a Cluster.

    A Cluster cannot simply contain VirtualIOServer links because it is
    likely that some of the Cluster's members are not managed by the same
    instance of the PowerVM REST server, which would then have no way to
    construct said links.  In such cases, the Node object supplies enough
    information about the VIOS that it could be found by a determined consumer.

    To add a new Node to a Cluster, only the hostname is required.
    n = Node()
    n.hostname = ...
    cluster.nodes.append(n)
    adapter.update(...)
    """

    @classmethod
    def bld(cls, adapter, hostname=None, lpar_id=None, mtms=None,
            vios_uri=None):
        """Create a fresh Node ElementWrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param hostname: String hostname (or IP) of the Node.
        :param lpar_id: Integer LPAR ID of the Node.
        :param mtms: String OR mtms.MTMS wrapper representing the
                     Machine Type, Model, and Serial Number of the system
                     hosting the VIOS.  String format: 'MT-M*S'
                     e.g. '8247-22L*1234A0B'.
        :param vios_uri: String URI representing this Node.
        """
        node = cls._bld(adapter)
        if vios_uri:
            node._vios_uri(vios_uri)
        if lpar_id:
            node._lpar_id(lpar_id)
        if mtms:
            node._mtms(mtms)
        if hostname:
            node._hostname(hostname)
        return node

    @property
    def hostname(self):
        return self._get_val_str(_N_HOSTNAME)

    def _hostname(self, hn):
        self.set_parm_value(_N_HOSTNAME, hn)

    @property
    def lpar_id(self):
        """Small integer partition ID, not UUID."""
        return self._get_val_int(_N_LPARID)

    def _lpar_id(self, new_lpar_id):
        self.set_parm_value(_N_LPARID, str(new_lpar_id))

    @property
    def mtms(self):
        """MTMS Element wrapper of the system hosting the Node (VIOS)."""
        return mtmwrap.MTMS.wrap(self._find(mtmwrap.MTMS_ROOT))

    def _mtms(self, new_mtms):
        """Sets the MTMS of the Node.

        :param new_mtms: May be either a string of the form 'MT-M*S' or a
                         mtms.MTMS ElementWrapper.
        """
        if not isinstance(new_mtms, mtmwrap.MTMS):
            new_mtms = mtmwrap.MTMS.bld(self.adapter, new_mtms)
        self.inject(new_mtms.element)

    @property
    def vios_uri(self):
        """The URI of the VIOS.

        This is only set if the VIOS is on this system!
        """
        return self.get_href(_N_VIOS_LINK, one_result=True)

    def _vios_uri(self, new_uri):
        self.set_href(_N_VIOS_LINK, new_uri)

    @property
    def vios_uuid(self):
        """The UUID of the Node (VIOS).

        This is only set if the VIOS is on this system!
        """
        uri = self.vios_uri
        if uri is not None:
            return u.get_req_path_uuid(uri, preserve_case=True)

    @property
    def state(self):
        return self._get_val_str(_N_STATE)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Specialized tasks for NPIV World-Wide Port Names (WWPNs)."""

from oslo_log import log as logging

from pypowervm import const as c
from pypowervm import exceptions as e
from pypowervm.i18n import _
from pypowervm import util as u
from pypowervm.utils import uuid
from pypowervm.wrappers import base_partition as bp
from pypowervm.wrappers import job as pvm_job
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import virtual_io_server as pvm_vios

import six

LOG = logging.getLogger(__name__)


_ANY_WWPN = '-1'
_FUSED_ANY_WWPN = '-1 -1'
_GET_NEXT_WWPNS = 'GetNextWWPNs'


def build_wwpn_pair(adapter, host_uuid, pair_count=1):
    """Builds a WWPN pair that can be used for a VirtualFCAdapter.

    Note: The API will only generate up to 8 pairs at a time.  Any more will
    cause the API to raise an error.

    :param adapter: The adapter to talk over the API.
    :param host_uuid: The host system for the generation.
    :param pair_count: (Optional, Default: 1) The number of WWPN pairs to
                       generate.  Can not be more than 8 or else the API will
                       fail.
    :return: Non-mutable WWPN Pairs (list)
    """
    # Build up the job & invoke
    resp = adapter.read(
        pvm_ms.System.schema_type, root_id=host_uuid,
        suffix_type=c.SUFFIX_TYPE_DO, suffix_parm=_GET_NEXT_WWPNS)
    job_w = pvm_job.Job.wrap(resp)
    job_p = [job_w.create_job_parameter('numberPairsRequested',
                                        str(pair_count))]
    job_w.run_job(host_uuid, job_parms=job_p)

    # Get the job result, and parse the output.
    job_result = job_w.get_job_results_as_dict()
    return job_result['wwpnList'].split(',')


def find_vios_for_wwpn(vios_wraps, p_port_wwpn):
    """Will find the VIOS that has a PhysFCPort for the p_port_wwpn.

    :param vios_wraps: A list or set of VIOS wrappers.
    :param p_port_wwpn: The physical port's WWPN.
    :return: The VIOS wrapper that contains a physical port with the WWPN.
             If there is not one, then None will be returned.
    :return: The port (which is a PhysFCPort wrapper) on the VIOS wrapper that
             represents the physical port.
    """
    # Sanitize our input
    s_p_port_wwpn = u.sanitize_wwpn_for_api(p_port_wwpn)
    for vios_w in vios_wraps:
        for port in vios_w.pfc_ports:
            # No need to sanitize the API WWPN, it comes from the API.
            if u.sanitize_wwpn_for_api(port.wwpn) == s_p_port_wwpn:
                return vios_w, port
    return None, None


def find_vios_for_vfc_wwpns(vios_wraps, vfc_wwpns):
    """Will find the VIOS that is hosting the vfc_wwpns.

    :param vios_wraps: A list or set of VIOS wrappers.
    :param vfc_wwpns: The list or set of virtual fibre channel WWPNs.
    :return: The VIOS wrapper that supports the vfc adapters.  If there is not
             one, then None will be returned.
    :return: The VFCMapping on the VIOS that supports the client adapters.
    """
    # Sanitize our input
    vfc_wwpns = {u.sanitize_wwpn_for_api(x) for x in vfc_wwpns}
    for vios_w in vios_wraps:
        for vfc_map in vios_w.vfc_mappings:
            # If the map has no client adapter...then move on
            if not vfc_map.client_adapter:
                continue

            # If the WWPNs match, return it
            if vfc_wwpns == set(vfc_map.client_adapter.wwpns):
                return vios_w, vfc_map
    return None, None


def intersect_wwpns(wwpn_set1, wwpn_set2):
    """Will return the intersection of WWPNs between the two sets.

    :param wwpn_set1: A list of WWPNs.
    :param wwpn_set2: A list of WWPNs.
    :return: The intersection of the WWPNs.  Will maintain the WWPN format
             of wwpn_set1, but the comparison done will be agnostic of
             formats (ex. colons and/or upper/lower case).
    """
    wwpn_set2 = [u.sanitize_wwpn_for_api(x) for x in wwpn_set2]
    return [y for y in wwpn_set1 if u.sanitize_wwpn_for_api(y) in wwpn_set2]


def derive_base_npiv_map(vios_wraps, p_port_wwpns, v_port_count):
    """Builds a blank NPIV port mapping, without any known vFC WWPNs.

    This method is functionally similar to the derive_npiv_map.  However, the
    derive_npiv_map method assumes knowledge of the Virtual Fibre Channel
    mappings beforehand.  This method will generate a similar map, but when
    sent to the add_map method, that method will allow the API to generate the
    globally unique WWPNs rather than pre-seeding them.

    :param vios_wraps: A list of VIOS wrappers.  Can be built using the
                       extended attribute group (xag) of VIO_FMAP.
    :param p_port_wwpns: A list of the WWPNs (strings) that can be used to
                         map the ports to.  These WWPNs should reside on
                         Physical FC Ports on the VIOS wrappers that were
                         passed in.
    :param v_port_count: The number of virtual ports to create.
    :return: A list of sets.  The format will be similar to that of the
             derive_npiv_map method.  However, instead of a fused_vfc_port_wwpn
             a marker will be used to indicate that the API should generate
             the WWPN.
    """
    # Double the count of the markers.  Should result in -1 -1 as the WWPN.
    v_port_markers = [_ANY_WWPN] * v_port_count * 2
    return derive_npiv_map(vios_wraps, p_port_wwpns, v_port_markers)


def derive_npiv_map(vios_wraps, p_port_wwpns, v_port_wwpns, preserve=True):
    """This method will derive a NPIV map.

    A NPIV map is the linkage between an NPIV virtual FC Port and the backing
    physical port.  Two v_port_wwpns get tied to an individual p_port_wwpn.

    A list of the 'mappings' will be returned.  One per pair of v_port_wwpns.

    The mappings will first attempt to spread across the VIOSes.  Within each
    VIOS, the port with the most available free NPIV ports will be selected.

    There are scenarios where ports on a single VIOS could be reused.
     - 4 v_port_wwpns, all p_port_wwpns reside on single VIOS
     - 8 v_port_wwpns, only two VIOSes
     - Etc...

    In these scenarios, the ports will be spread such that they're running
    across all the physical ports (that were passed in) on a given VIOS.

    In even rarer scenarios, the same physical port may be re-used if the
    v_port_wwpn pairs exceed the total number of p_port_wwpns.

    :param vios_wraps: A list of VIOS wrappers.  Can be built using the
                       extended attribute group (xag) of VIO_FMAP.
    :param p_port_wwpns: A list of the WWPNs (strings) that can be used to
                         map the ports to.  These WWPNs should reside on
                         Physical FC Ports on the VIOS wrappers that were
                         passed in.
    :param v_port_wwpns: A list of the virtual fibre channel port WWPNs.  Must
                         be an even number of ports.
    :param preserve: (Optional, Default=True) If True, existing mappings with
                     matching virtual fibre channel ports are preserved. Else
                     new mappings are generated.
    :return: A list of tuples representing both new and preserved mappings.
             The format will be:
      [ (p_port_wwpn1, fused_vfc_port_wwpn1),
        (p_port_wwpn2, fused_vfc_port_wwpn2),
        etc... ]

    A 'fused_vfc_port_wwpn' is simply taking two v_port_wwpns, sanitizing them
    and then putting them into a single string separated by a space.
    """

    # Fuse all the v_port_wwpns together.
    fused_v_port_wwpns = _fuse_vfc_ports(v_port_wwpns)

    # Up front sanitization of all the p_port_wwpns
    p_port_wwpns = list(map(u.sanitize_wwpn_for_api, p_port_wwpns))

    existing_maps = []
    new_fused_wwpns = []

    # Detect if any mappings already exist on the system.
    for fused_v_wwpn in fused_v_port_wwpns:
        # If the mapping already exists, then add it to the existing maps.
        vfc_map = has_client_wwpns(vios_wraps, fused_v_wwpn.split(" "))[1]
        # Preserve an existing mapping if preserve=True. Otherwise, the
        # backing_port may not be set and this is not an error condition if
        # the vfc mapping is getting rebuilt.
        if vfc_map is not None and preserve:
            mapping = (vfc_map.backing_port.wwpn, fused_v_wwpn)
            existing_maps.append(mapping)
        else:
            new_fused_wwpns.append(fused_v_wwpn)
            LOG.debug("Add new map for client wwpns %s. Existing map=%s, "
                      "preserve=%s", fused_v_wwpn, vfc_map, preserve)

    # Determine how many mappings are needed.
    needed_maps = len(new_fused_wwpns)
    newly_built_maps = []

    next_vio_pos = 0
    fuse_map_pos = 0
    loops_since_last_add = 0

    # This loop will continue through each VIOS (first set of load balancing
    # should be done by VIOS) and if there are ports on that VIOS, will add
    # them to the mapping.
    #
    # There is a rate limiter here though.  If none of the VIOSes are servicing
    # the request, then this has potential to be infinite loop.  The rate
    # limiter detects such a scenario and will prevent it from occurring.  In
    # these cases the UnableToFindFCPortMap exception is raised.
    #
    # As such, limit it such that if no VIOS services the request, we break
    # out of the loop and throw error.
    while len(newly_built_maps) < needed_maps:
        # Walk through each VIOS.
        vio = vios_wraps[next_vio_pos]
        loops_since_last_add += 1

        # If we've been looping more than the VIOS count, we need to exit.
        # Something has gone amuck.
        if loops_since_last_add > len(vios_wraps):
            raise e.UnableToFindFCPortMap()

        # This increments the VIOS position for the next loop
        next_vio_pos = (next_vio_pos + 1) % len(vios_wraps)

        # Find the FC Ports that are on this system.
        potential_ports = _find_ports_on_vio(vio, p_port_wwpns)
        if len(potential_ports) == 0:
            # No ports on this VIOS.  Continue to next.
            continue

        # Next, from the potential ports, find the PhysFCPort that we should
        # use for the mapping.
        new_map_port = _find_map_port(potential_ports,
                                      newly_built_maps + existing_maps)
        if new_map_port is None:
            # If there was no mapping port, then we should continue on to
            # the next VIOS.
            continue

        # Add the mapping!
        mapping = (new_map_port.wwpn, new_fused_wwpns[fuse_map_pos])
        fuse_map_pos += 1
        newly_built_maps.append(mapping)
        loops_since_last_add = 0

    # Mesh together the existing mapping lists plus the newly built ports.
    return newly_built_maps + existing_maps


def _find_map_port(potential_ports, mappings):
    """Will determine which port to use for a new mapping.

    :param potential_ports: List of PhysFCPort wrappers that are candidate
                            ports.
    :param mappings: The existing mappings, as generated by derive_npiv_map.
    :return: The PhysFCPort that should be used for this mapping.
    """
    # The first thing that we need is to understand how many physical ports
    # have been used by mappings already.  This is important for the scenarios
    # where we're looping across the same set of physical ports on this VIOS.
    # We should avoid reusing those same physical ports as our previous
    # mappings as much as possible, to allow for as much physical multi pathing
    # as possible.
    #
    # This dictionary will have a key of every port's WWPN, within that it will
    # have a dictionary which contains 'port_mapping_use' and then the 'port'
    # which is the port passed in.
    port_dict = dict((p.wwpn, {'port_mapping_use': 0, 'port': p})
                     for p in potential_ports)

    for mapping in mappings:
        p_wwpn = mapping[0]

        # If this physical WWPN is not in our port_dict, then we know that
        # this port is not on the VIOS.  Therefore, we can't take it into
        # consideration.
        if p_wwpn not in port_dict.keys():
            continue

        # Increment our counter to indicate that a previous mapping already
        # has used this physical port.
        port_dict[p_wwpn]['port_mapping_use'] += 1

    # Now find the set of ports with the lowest count of usage by mappings.
    # The first time through this method, this will be all the physical ports.
    # This is only interesting once the previous mappings come into play.
    # This is where we reduce the 'candidate physical ports' down to those
    # that are least used by our existing mappings.
    #
    # There is a reasonable upper limit of 128 mappings (which should be
    # beyond what admins will want to map vFC's to a single pFC).  We simply
    # put that in to avoid infinite loops in the extremely, almost unimaginable
    # event that we've reached an upper boundary.  :-)
    #
    # Subsequent logic should be OK with this, as we simply will return None
    # for the next available port.
    starting_count = 0
    list_of_cand_ports = []
    while len(list_of_cand_ports) == 0 and starting_count < 128:
        for port_info in port_dict.values():
            if port_info['port_mapping_use'] == starting_count:
                list_of_cand_ports.append(port_info['port'])

        # Increment the count, in case we have to loop again.
        starting_count += 1

    # At this point, the list_of_cand_ports is essentially a list of ports
    # least used by THIS mapping.  Now, we need to narrow that down to the
    # port that has the most npiv_available_ports.  The one with the most
    # available ports is the least used.  Therefore the best candidate (that
    # we can choose with limited info).
    high_avail_port = None
    for port in list_of_cand_ports:
        # If this is the first port, or this new port has more available NPIV
        # slots, then use that.
        if (high_avail_port is None or
                high_avail_port.npiv_available_ports <
                port.npiv_available_ports):
            high_avail_port = port

    return high_avail_port


def _find_ports_on_vio(vio_w, p_port_wwpns):
    """Will return a list of Physical FC Ports on the vio_w.

    :param vio_w: The VIOS wrapper.
    :param p_port_wwpns: The list of all physical ports.  May exceed the
                         ports on the VIOS.
    :return: List of the physical FC Port wrappers that are on the VIOS
             for the WWPNs that exist on this system.
    """
    return [port for port in vio_w.pfc_ports
            if u.sanitize_wwpn_for_api(port.wwpn) in p_port_wwpns]


def _fuse_vfc_ports(wwpn_list):
    """Returns a list of fused VFC WWPNs.  See derive_npiv_map."""
    l = list(map(u.sanitize_wwpn_for_api, wwpn_list))
    return list(map(' '.join, zip(l[::2], l[1::2])))


def _find_pfc_wwpn_by_name(vios_w, pfc_name):
    """Returns the physical port wwpn within a VIOS based off the FC port name.

    :param vios_w: VIOS wrapper.
    :param pfc_name: The physical fibre channel port name.
    """
    for port in vios_w.pfc_ports:
        if port.name == pfc_name:
            return port.wwpn
    return None


def find_maps(mapping_list, client_lpar_id, client_adpt=None, port_map=None):
    """Filter a list of VFC mappings by LPAR ID.

    This is based on scsi_mapper.find_maps, but does not yet provide all the
    same functionality.

    :param mapping_list: The mappings to filter.  Iterable of VFCMapping.
    :param client_lpar_id: Integer short ID or string UUID of the LPAR on the
                           client side of the mapping.  Note that the UUID form
                           relies on the presence of the client_lpar_href
                           field.  Some mappings lack this field, and would
                           therefore be ignored.
    :param client_adpt: (Optional, Default=None) If set, will only include the
                        mapping if the client adapter's WWPNs match as well.
    :param port_map: (Optional, Default=None) If set, will look for a matching
                     mapping based off the client WWPNs as specified by the
                     port mapping.  The format of this is defined by the
                     derive_npiv_map method.
    :return: A list comprising the subset of the input mapping_list whose
             client LPAR IDs match client_lpar_id.
    """
    is_uuid, client_id = uuid.id_or_uuid(client_lpar_id)
    matching_maps = []

    if port_map:
        v_wwpns = [u.sanitize_wwpn_for_api(x) for x in port_map[1].split()]

    for vfc_map in mapping_list:
        # If to a different VM, continue on.
        href = vfc_map.client_lpar_href
        if is_uuid and (not href or client_id != u.get_req_path_uuid(
                href, preserve_case=True)):
            continue
        elif not is_uuid and vfc_map.server_adapter.lpar_id != client_id:
            # Use the server adapter ^^ in case this is an orphan.
            continue

        # If there is a client adapter, and it is not a 'ANY WWPN', then
        # check to see if the mappings match.
        if client_adpt and client_adpt.wwpns != {_ANY_WWPN}:
            # If they passed in a client adapter, but the map doesn't have
            # one, then we have to ignore
            if not vfc_map.client_adapter:
                continue

            # Check to make sure the WWPNs between the two match.  This should
            # be an order independence check (as this query shouldn't care...
            # but the API itself does care about order).
            if set(client_adpt.wwpns) != set(vfc_map.client_adapter.wwpns):
                continue

        # If the user had a port map, do the virtual WWPNs from that port
        # map match the client adapter wwpn map.
        if port_map:
            if vfc_map.client_adapter is None:
                continue

            # If it is a new mapping with generated WWPNs, then the client
            # adapter can't have WWPNs.
            if v_wwpns == [_ANY_WWPN, _ANY_WWPN]:
                if vfc_map.client_adapter.wwpns != []:
                    continue
            elif set(vfc_map.client_adapter.wwpns) != set(v_wwpns):
                continue

        # Found a match!
        matching_maps.append(vfc_map)

    return matching_maps


def remove_maps(v_wrap, client_lpar_id, client_adpt=None, port_map=None):
    """Remove one or more VFC mappings from a VIOS wrapper.

    The changes are not flushed back to the REST server.

    :param v_wrap: VIOS EntryWrapper representing the Virtual I/O Server whose
                   VFC mappings are to be updated.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param client_adpt: (Optional, Default=None) If set, will only add the
                        mapping if the client adapter's WWPNs match as well.
    :param port_map: (Optional, Default=None) If set, will look for a matching
                     mapping based off the client WWPNs as specified by the
                     port mapping.  The format of this is defined by the
                     derive_npiv_map method.
    :return: The mappings removed from the VIOS wrapper.
    """
    resp_list = []

    for matching_map in find_maps(v_wrap.vfc_mappings, client_lpar_id,
                                  client_adpt=client_adpt,
                                  port_map=port_map):
        v_wrap.vfc_mappings.remove(matching_map)
        resp_list.append(matching_map)
    return resp_list


def find_vios_for_port_map(vios_wraps, port_map):
    """Finds the appropriate VIOS wrapper for a given port map.

    Note that the algorithm first checks based off of the client WWPNs.  If the
    client WWPNs can not be found (perhaps the map is still -1 -1 from the
    derive_base_npiv_map) then the physical port WWPN will be checked.

    :param vios_wraps: A list of Virtual I/O Server wrapper objects.
    :param port_map: The port mapping (as defined by the derive_npiv_map
                     method).
    :return: The Virtual I/O Server wrapper that supports the port map.
    """
    # Check first based off the client WWPNs.  Note that this may be -1 -1
    # in which case it will return nothing
    vios_w = find_vios_for_vfc_wwpns(vios_wraps, port_map[1].split())[0]
    if vios_w:
        return vios_w

    # If we had nothing, check based off the physical port WWPN.  The
    # reason this is not the first check is because the mapping may be mid
    # live migration, thus pointing to a source.  But if that was the case
    # then the first check would have returned the right WWPNs.  The only
    # time this should be hit is in the middle of a create operation.
    return find_vios_for_wwpn(vios_wraps, port_map[0])[0]


def add_map(vios_w, host_uuid, lpar_uuid, port_map, error_if_invalid=True,
            lpar_slot_num=None):
    """Adds a vFC mapping to a given VIOS wrapper.

    These changes are not flushed back to the REST server.  The wrapper itself
    is simply modified.

    :param vios_w: VIOS EntryWrapper representing the Virtual I/O Server whose
                   VFC mappings are to be updated.
    :param host_uuid: The pypowervm UUID of the host.
    :param lpar_uuid: The pypowervm UUID of the client LPAR to attach to.
    :param port_map: The port mapping (as defined by the derive_npiv_map
                     method).
    :param error_if_invalid: (Optional, Default: True) If the port mapping
                             physical port can not be found, raise an error.
    :param lpar_slot_num: (Optional, Default: None) The client adapter
                          VirtualSlotNumber to be set. If None the next
                          available slot would be used.
    :return: The VFCMapping that was added or updated with a missing backing
             port.  If the mapping already existed then None is returned.
    """
    # This is meant to find the physical port.  Can run against a single
    # element.  We assume invoker has passed correct VIOS.
    new_vios_w, p_port = find_vios_for_wwpn([vios_w], port_map[0])
    if new_vios_w is None:
        if error_if_invalid:
            # Log the payload in the response.
            LOG.warning(_("Unable to find appropriate VIOS.  The payload "
                          "provided was likely insufficient.  The payload "
                          "data is:\n %s)"), vios_w.toxmlstring(pretty=True))
            raise e.UnableToDerivePhysicalPortForNPIV(wwpn=port_map[0],
                                                      vio_uri=vios_w.href)
        else:
            return None

    v_wwpns = None
    if port_map[1] != _FUSED_ANY_WWPN:
        v_wwpns = [u.sanitize_wwpn_for_api(x) for x in port_map[1].split()]

    if v_wwpns is not None:
        for vfc_map in vios_w.vfc_mappings:
            if (vfc_map.client_adapter is None or
                    vfc_map.client_adapter.wwpns is None):
                continue
            if set(vfc_map.client_adapter.wwpns) != set(v_wwpns):
                continue

            # If we reach this point, we know that we have a matching map.
            # Check that the physical port is set in the mapping.
            if vfc_map.backing_port:
                LOG.debug("Matching existing vfc map found with backing port:"
                          " %s", vfc_map.backing_port.wwpn)
                # The attach of this volume, for this vFC mapping is complete.
                # Nothing else needs to be done, exit the method.
                return None
            else:
                LOG.info(_("The matched VFC port map has no backing port set."
                           " Adding %(port)s to mapping for client wwpns: "
                           "%(wwpns)s"),
                         {'port': p_port.name, 'wwpns': v_wwpns})
                # Build the backing_port and add it to the vfc_map.
                vfc_map.backing_port = bp.PhysFCPort.bld_ref(
                    vios_w.adapter, p_port.name, ref_tag='Port')
                return vfc_map

    # However, if we hit here, then we need to create a new mapping and
    # attach it to the VIOS mapping
    vfc_map = pvm_vios.VFCMapping.bld(vios_w.adapter, host_uuid, lpar_uuid,
                                      p_port.name, client_wwpns=v_wwpns,
                                      lpar_slot_num=lpar_slot_num)
    vios_w.vfc_mappings.append(vfc_map)
    return vfc_map


def has_client_wwpns(vios_wraps, client_wwpn_pair):
    """Returns the vios wrapper and vfc map if the client WWPNs already exist.

    :param vios_wraps: The VIOS wrappers.  Should be queried with the
                       VIO_FMAP extended attribute.
    :param client_wwpn_pair: The pair (list or set) of the client WWPNs.
    :return vios_w: The VIOS wrapper containing the wwpn pair.  None if none
                    of the wrappers contain the pair.
    :return vfc_map: The mapping containing the client pair.  May be None.
    """
    client_wwpn_pair = set([u.sanitize_wwpn_for_api(x)
                            for x in client_wwpn_pair])

    for vios_wrap in vios_wraps:
        for vfc_map in vios_wrap.vfc_mappings:
            if vfc_map.client_adapter is None:
                continue

            pair = set([u.sanitize_wwpn_for_api(x)
                        for x in vfc_map.client_adapter.wwpns])
            if pair == client_wwpn_pair:
                return vios_wrap, vfc_map

    return None, None


def build_migration_mappings_for_fabric(vios_wraps, p_port_wwpns,
                                        client_slots):
    """Builds the vFC migration mappings for a given fabric.

    This method will build the migration mappings for a given fabric.
    The response is a list of strings that can be used in the migration.py

    Note: If you have multiple fabrics, then each fabric will need to
    independently call this method with the appropriate p_port_wwpns.

    Note: This must be run on the destination server before the migration.
    It is typically input back to the source server for the migration call.

    :param vios_wraps: The VIOS wrappers for the target system.  Must
                       have the VIO_FMAP xag specified.
    :param p_port_wwpns: The physical port WWPNs that can be used for
                         this specific fabric.  May span multiple VIOSes,
                         but each must be part of the vios_wraps.
    :param client_slots: A list of integers which represent the *source*
                         system's LPAR virtual Fibre Channel slots that
                         are participating in this fabric.
    :return: List of mappings that can be passed into the migration.py
             for the live migration.  The format is defined within the
             migration.py, migrate_lpar method.
    """
    basic_mappings = derive_base_npiv_map(vios_wraps, p_port_wwpns,
                                          len(client_slots))
    resp = []
    for basic_map, client_slot in zip(basic_mappings, client_slots):
        # Find the appropriate VIOS hosting this physical port.
        vios_w, port = find_vios_for_wwpn(vios_wraps, basic_map[0])

        # The format is:
        #     virtual-slot-number/vios-lpar-name/vios-lpar-ID
        #     [/[vios-virtual-slot-number][/[vios-fc-port-name]]]
        #
        # We do not specify the vios-virtual-slot-number.
        resp.append(str(client_slot) + "/" + vios_w.name + "/" +
                    str(vios_w.id) + "//" + port.name)
    return resp


def _split_ports_per_fabric(slot_grouping, fabric_data):
    """Splits the slots per fabric which are to be placed on the same VIOS.

    :param slot_grouping: The slots which are to be placed in the same vios
                          Ex:
                          [3, 6]
                          Here the slots 3 and 6 are to be placed on the same
                          vios.
    :param fabric_data: Dictionary where the key is the fabric name.  The
                        value is another dictionary with the slots and the
                        p_port_wwpns.

                        Ex:
                        { 'A': {'slots': [3, 4, 5], p_port_wwpns: [1, 2, 3] },
                          'B': {'slots': [6, 7, 8], p_port_wwpns: [4, 5] } }

                        The slot indicates which slots from the client slots
                        align with which fabric.
    :return resp: The slots which can be placed on the same VIOS alone are
                  returned.
                  {'A': {'slots': [3],  p_port_wwpns: [1, 2, 3] },
                   'B': {'slots': [6],  p_port_wwpns: [4, 5] } }
    """
    resp = {}
    for fabric in fabric_data:
        slots = [x for x in fabric_data[fabric]['slots'] if x in slot_grouping]
        if not slots:
            continue

        resp[fabric] = {'slots': slots,
                        'p_port_wwpns': fabric_data[fabric]['p_port_wwpns']}
    return resp


def _does_vios_support_split_map(vios_w, split_map):
    """Split_map provided by _split_ports_per_fabric.

    :param vios_w: The VIOS wrapper to validate if the split map can match
    :param split_map: This contains the physical ports and the slots to be
                      matched on the VIOS
                      Ex: {'A': {'p_port_wwpns': ['phy_port1'], 'slots': [3]}
                      Check if the number of slots required can be satisfied
                      by the given VIOS.
    """
    for fabric_map in six.itervalues(split_map):
        needed_pports = len(fabric_map['slots'])
        fabric_pports_on_vios = _find_ports_on_vio(
            vios_w, fabric_map['p_port_wwpns'])
        if len(fabric_pports_on_vios) < needed_pports:
            return False
    return True


def build_migration_mappings(vios_wraps, fabric_data, slot_peers):
    """Builds the vFC migration mappings.

    Looks holistically at the system.  Should generally be used instead of
    build_migration_mappings_for_fabric.

    :param vios_wraps: The VIOS wrappers for the target system.  Must
                       have the VIO_FMAP xag specified.
    :param fabric_data: Dictionary where the key is the fabric name.  The
                        value is another dictionary with the slots and the
                        p_port_wwpns.

                        Ex:
                        { 'A': {'slots': [3, 4, 5], p_port_wwpns: [1, 2, 3] },
                          'B': {'slots': [6, 7, 8], p_port_wwpns: [4, 5] } }

                        The slot indicates which slots from the client slots
                        align with which fabric.

    :param slot_peers: An array of arrays.  Indicates all of the slots that
                       need to be grouped together on a given VIOS.

                       Ex:
                       [ [3, 6, 7], [4, 8], [5] ]

                       Indicates that (based on the data from fabric_data)
                       one VIOS must host client slot 3 (from fabric A) and
                       slots 6 and 7 (from fabric B).  Then another VIOS must
                       host client slot 4 (from fabric A) and client slot 8
                       (from fabric B).  And the third VIOS must only host
                       client slot 5 (from fabric A).
    :return: List of mappings that can be passed into the migration.py
             for the live migration.  The format is defined within the
             migration.py, migrate_lpar method.
    """

    # First sort the slot peers.  The one with the most peers needs to go first
    # then work down from there.
    slot_peers = sorted(slot_peers, key=len, reverse=True)

    vios_to_split_map = {}

    # We create a map of all the VIOSes and their corresponding slots
    for peer_grouping in slot_peers:
        split_map = _split_ports_per_fabric(peer_grouping, fabric_data)
        LOG.debug("split_map %s" % split_map)
        for vios_w in vios_wraps:
            LOG.debug("Checking vios name %(name)s vios fc ports %(port)s" %
                      dict(name=vios_w.name,
                           port=[port.wwpn for port in vios_w.pfc_ports]))
            if vios_w in vios_to_split_map.keys():
                continue
            if not _does_vios_support_split_map(vios_w, split_map):
                continue
            vios_to_split_map[vios_w] = split_map
            break
        else:
            # When no vios match is found for peer group error
            raise e.UnableToFindFCPortMap()

    LOG.debug("vios_to_split_map %s" % vios_to_split_map)
    resp = []
    # Each VIOS has a split map.  the split map contains the fabric (as
    # the key), the physical port wwpns, and the slots.
    for vios_w, split_map in six.iteritems(vios_to_split_map):
        for fabric_map in six.itervalues(split_map):
            p_port_wwpns = fabric_map['p_port_wwpns']
            slots = fabric_map['slots']
            basic_mappings = derive_base_npiv_map([vios_w], p_port_wwpns,
                                                  len(slots))
            for basic_map, client_slot in zip(basic_mappings, slots):
                # Find the appropriate VIOS hosting this physical port.
                vios_w, port = find_vios_for_wwpn([vios_w], basic_map[0])

                # The format is:
                #     virtual-slot-number/vios-lpar-name/vios-lpar-ID
                #     [/[vios-virtual-slot-number][/[vios-fc-port-name]]]
                #
                # We do not specify the vios-virtual-slot-number.
                items = (str(client_slot), vios_w.name, str(vios_w.id),
                         '', port.name)
                resp.append("/".join(items))
    return resp



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

import pytz

import pypowervm.tests.test_utils.test_wrapper_abc as twrap
from pypowervm.wrappers import monitor

_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f%Z'


class TestPCM(twrap.TestWrapper):

    file = 'pcm_pref.txt'
    wrapper_class_to_test = monitor.PcmPref

    def test_pcm(self):
        pcm_wrap = self.entries[0]
        self.assertEqual('dev-system-6', pcm_wrap.system_name)

        # Test enable getters when all are loaded and False
        self.assertFalse(pcm_wrap.ltm_enabled)
        self.assertFalse(pcm_wrap.aggregation_enabled)
        self.assertFalse(pcm_wrap.stm_enabled)
        self.assertFalse(pcm_wrap.compute_ltm_enabled)

        # Set all to True and test.
        pcm_wrap.ltm_enabled = True
        self.assertTrue(pcm_wrap.ltm_enabled)

        pcm_wrap.aggregation_enabled = True
        self.assertTrue(pcm_wrap.aggregation_enabled)

        pcm_wrap.stm_enabled = True
        self.assertTrue(pcm_wrap.stm_enabled)

        pcm_wrap.compute_ltm_enabled = True
        self.assertTrue(pcm_wrap.compute_ltm_enabled)

    def test_str_to_datetime(self):
        expected = datetime.datetime(
            year=2015, month=4, day=30, hour=6, minute=11,
            second=35).replace(tzinfo=pytz.utc)
        func = monitor.MonitorMetrics._str_to_datetime

        self.assertEqual(expected, func('2015-04-30T11:11:35.000-05:00'))
        self.assertEqual(expected, func('2015-04-30T01:11:35.000+05:00'))
        self.assertEqual(expected, func('2015-04-30T06:11:35.000-00:00'))
        self.assertEqual(expected, func('2015-04-30T06:11:35.000Z'))


class TestLTMMetrics(twrap.TestWrapper):

    file = 'ltm_feed.txt'
    wrapper_class_to_test = monitor.LTMMetrics

    def test_ltm_metrics(self):
        link = ('https://9.1.2.3:12443/rest/api/pcm/ManagedSystem/98498bed'
                '-c78a-3a4f-b90a-4b715418fcb6/RawMetrics/LongTermMonitor/L'
                'TM_8247-22L*1111111_vios_2_20150430T035300+0000.json')

        wrap = self.entries[0]
        self.assertEqual('15161241-b72f-41d5-8154-557ff699fb75', wrap.id)
        self.assertEqual('2015-04-30T03:53:00.000Z', wrap.published)
        self.assertEqual(
            'LTM_8247-22L*1111111_vios_2_20150430T035300+0000.json',
            wrap.title)
        self.assertEqual('2015-04-30T03:53:00.000Z', wrap.updated)
        self.assertEqual('vios_2', wrap.category)
        self.assertEqual(link, wrap.link)

        # Test wrapping just one entry and we should get the same data
        wrap = monitor.LTMMetrics.wrap(self.entries[0].entry)
        self.assertEqual(link, wrap.link)


class TestSTMMetrics(twrap.TestWrapper):

    file = 'stm_feed.txt'
    wrapper_class_to_test = monitor.STMMetrics

    def test_stm_metrics(self):
        link = ('https://9.1.2.3:12443/rest/api/pcm/ManagedSystem/98498bed'
                '-c78a-3a4f-b90a-4b715418fcb6/RawMetrics/ShortTermMonitor/'
                'STM_8247-22L*1111111_phyp_20150430T061135+0000.json')

        wrap = self.entries[0]
        self.assertEqual('28cb2328-ca14-48ef-a3bd-691debef53dd', wrap.id)
        self.assertEqual('2015-04-30T06:11:35.000-05:00', wrap.published)
        self.assertEqual('2015-04-30T01:11:35.000000UTC',
                         wrap.published_datetime.strftime(_DATETIME_FORMAT))
        self.assertEqual(
            'STM_8247-22L*1111111_phyp_20150430T061135+0000.json',
            wrap.title)
        self.assertEqual('2015-04-30T06:11:35.002Z', wrap.updated)
        self.assertEqual('2015-04-30T06:11:35.002000UTC',
                         wrap.updated_datetime.strftime(_DATETIME_FORMAT))
        self.assertEqual('phyp', wrap.category)
        self.assertEqual(link, wrap.link)

        # Test wrapping just one entry and we should get the same data
        wrap = monitor.STMMetrics.wrap(self.entries[0].entry)
        self.assertEqual(link, wrap.link)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Complex tasks around SR-IOV cards/ports, VFs, and vNICs."""

import copy
from oslo_concurrency import lockutils as lock
from oslo_log import log as logging
import random
import six

import pypowervm.exceptions as ex
from pypowervm.i18n import _
import pypowervm.tasks.partition as tpar
import pypowervm.utils.transaction as tx
import pypowervm.wrappers.iocard as card
import pypowervm.wrappers.managed_system as ms

LOG = logging.getLogger(__name__)

# Take read_lock on operations that create/delete VFs (including VNIC).  This
# is a read_lock so we don't serialize all VF creation globally.
# Take write_lock on operations that modify properties of physical ports and
# rely on knowing the usage counts thereon (e.g. changing port labels).
PPORT_MOD_LOCK = lock.ReaderWriterLock()


def set_vnic_back_devs(vnic_w, pports, sys_w=None, vioses=None, redundancy=1,
                       capacity=None, check_port_status=False):
    """Set a vNIC's backing devices over given SRIOV physical ports and VIOSes.

    Assign the backing devices to a iocard.VNIC wrapper using an anti-affinity
    algorithm.  That is, the method attempts to distribute the backing devices
    across as diverse a range of physical SRIOV adapters and VIOSes as
    possible, using the least-saturated ports first.  For example, given:

    vios1, vios2

    SRIOVAdapter1
        PPortA (50% allocated)
        PPortB (20%)
        PPortC (45%)
    SRIOVAdapter2
        PPortD (10%)
        PPortE (2%)
        PPortF (11%)

    set_vnic_back_devs(vnic, [PPortA, PPortB, PPortC, PPortD, PPortE, PPortF],
                       [vios1, vios2], redundancy=4)

    ...we will create backing devices like:

    [(vios1, PPortE), (vios2, PPortB), (vios1, PPortD), (vios2, PPortC)]

    As part of the algorithm, we will use sriov_adaps to filter out physical
    ports which are already saturated.  This could err either way due to
    out-of-band changes:
    - We may end up excluding a port which has had some capacity freed up since
      sriov_adaps was retrieved; or
    - We may attempt to include a port which has become saturated since
      sriov_adaps was retrieved, resulting in an error from the REST server.

    This method acts on the vNIC-related capabilities on the system and VIOSes:
    - If the system is not vNIC capable, the method will fail.
    - If none of the active VIOSes are vNIC capable, the method will fail.
    - If redundancy > 1,
        - the system must be vNIC failover capable, and
        - at least one active VIOS must be vNIC failover capable.
    - If any VIOSes are vNIC failover capable, failover-incapable VIOSes will
      be ignored.

    :param vnic_w: iocard.VNIC wrapper, as created via VNIC.bld().  If
                   vnic_w.back_devs is nonempty, it is cleared and replaced.
                   This parameter is modified by the method (there is no return
                   value).  If this method raises an exception, vnic_w is
                   guaranteed to be unchanged.
    :param pports: List of physical location code strings (corresponding to the
                   loc_code @property of iocard.SRIOV*PPort) for all SRIOV
                   physical ports to be considered as backing devices for the
                   vNIC.  This does not mean that all of these ports will be
                   used.
    :param sys_w: Pre-fetched pypowervm.wrappers.managed_system.System wrapper.
                  If not specified, it will be fetched from the server.
    :param vioses: List of VIOS wrappers to consider for distribution of vNIC
                   servers.  Not all listed VIOSes will necessarily be used.
                   If not specified, the feed of all active (including RMC)
                   VIOSes will be fetched from the server.  If specified, the
                   list will be filtered to include only active (including RMC)
                   VIOSes (according to the wrappers - the server is not re-
                   checked).  The list is also filtered to remove VIOSes which
                   are not vNIC capable; and, if min_redundancy > 1, to remove
                   VIOSes which are not vNIC failover capable.
    :param redundancy: Number of backing devices to assign.  If the method
                       can't allocate this many VFs after filtering the pports
                       list, InsufficientSRIOVCapacity will be raised.  Note
                       that at most one VF is created on each physical port.
    :param capacity: (float) Minimum capacity to assign to each backing device.
                     Must be between 0.0 and 1.0, and must be a multiple of the
                     min_granularity of *all* of the pports.  (Capacity may be
                     assigned to each individual backing device after the fact
                     to achieve more control; but in that case, the consumer is
                     responsible for validating sufficient available capacity.)
    :param check_port_status: If True, only ports with link-up status will be
                              considered for allocation.  If False (the
                              default), link-down ports may be used.
    :raise NoRunningSharedSriovAdapters: If no SR-IOV adapters in Sriov mode
                                         and Running state can be found.
    :raise NotEnoughActiveVioses: If no active (including RMC) VIOSes can be
                                  found.
    :raise InsufficientSRIOVCapacity: If the method was not able to allocate
                                      enough VFs to satisfy the specified
                                      redundancy.
    :raise SystemNotVNICCapable: If the managed system is not vNIC capable.
    :raise NoVNICCapableVIOSes: If there are no vNIC-capable VIOSes.
    :raise VNICFailoverNotSupportedSys: If redundancy > 1, and the system is
                                        not vNIC failover capable.
    :raise VNICFailoverNotSupportedVIOS: If redundancy > 1, and there are no
                                         vNIC failover-capable VIOSes.
    """
    # An Adapter to work with
    adap = vnic_w.adapter
    if adap is None:
        raise ValueError('Developer error: Must build vnic_w with an Adapter.')

    # Check vNIC capability on the system
    sys_w = _check_sys_vnic_capabilities(adap, sys_w, redundancy)

    # Filter SR-IOV adapters
    sriov_adaps = _get_good_sriovs(sys_w.asio_config.sriov_adapters)

    # Get VIOSes which are a) active, b) vNIC capable, and c) vNIC failover
    # capable, if necessary.
    vioses = _check_and_filter_vioses(adap, vioses, redundancy)

    # Try not to end up lopsided on one VIOS
    random.shuffle(vioses)

    # Get the subset of backing ports corresponding to the specified location
    # codes which have enough space for new VFs.
    pport_wraps = _get_good_pport_list(sriov_adaps, pports, capacity,
                                       redundancy, check_port_status)

    # At this point, we've validated enough that we won't raise.  Start by
    # clearing any existing backing devices.
    vnic_w.back_devs = []

    card_use = {}
    for pport in pport_wraps:
        said = pport.sriov_adap_id
        if said not in card_use:
            card_use[said] = {'num_used': 0, 'ports_left': 0}
        card_use[said]['ports_left'] += 1
    vio_idx = 0
    while pport_wraps and len(vnic_w.back_devs) < redundancy:
        # Always rotate VIOSes
        vio = vioses[vio_idx]
        vio_idx = (vio_idx + 1) % len(vioses)
        # Select the least-saturated port from among the least-used adapters.
        least_uses = min([cud['num_used'] for cud in card_use.values()])
        pp2use = min([pport for pport in pport_wraps if
                      card_use[pport.sriov_adap_id]['num_used'] == least_uses],
                     key=lambda pp: pp.allocated_capacity)
        said = pp2use.sriov_adap_id
        # Register a hit on the chosen port's card
        card_use[said]['num_used'] += 1
        # And take off a port
        card_use[said]['ports_left'] -= 1
        # If that was the last port, remove this card from consideration
        if card_use[said]['ports_left'] == 0:
            del card_use[said]
        # Create and add the backing device
        vnic_w.back_devs.append(card.VNICBackDev.bld(
            adap, vio.uuid, said, pp2use.port_id, capacity=capacity))
        # Remove the port we just used from subsequent consideration.
        pport_wraps.remove(pp2use)


def _check_sys_vnic_capabilities(adap, sys_w, redundancy):
    """Validate vNIC capabilities on the Managed System.

    :param adap: pypowervm Adapter.
    :param sys_w: pypowervm.wrappers.managed_system.System wrapper.  If None,
                  it is retrieved from the host.
    :param redundancy: If greater than 1, this method will verify that the
                       System is vNIC failover-capable.  Otherwise, this check
                       is skipped.
    :return: The System wrapper.
    :raise SystemNotVNICCapable: If the System is not vNIC capable.
    :raise VNICFailoverNotSupportedSys: If min_redundancy > 1 and the System is
                                        not vNIC failover capable.
    """
    if sys_w is None:
        sys_w = ms.System.get(adap)[0]

    caps = sys_w.get_capabilities()
    if not caps.get('vnic_capable'):
        raise ex.SystemNotVNICCapable()
    if redundancy > 1 and not caps.get('vnic_failover_capable'):
        raise ex.VNICFailoverNotSupportedSys(red=redundancy)

    return sys_w


def _check_and_filter_vioses(adap, vioses, redundancy):
    """Return active VIOSes with appropriate vNIC capabilities.

    Remove all VIOSes which are not active or not vNIC capable.  If
    min_redundancy > 1, failover is required, so remove VIOSes that are not
    also vNIC failover capable.  Error if no VIOSes remain.

    :param adap: pypowervm Adapter.
    :param vioses: List of pypowervm.wrappers.virtual_io_server.VIOS to check.
                   If None, all active VIOSes are retrieved from the server.
    :param redundancy: If greater than 1, the return list will include only
                       vNIC failover-capable VIOSes.  Otherwise, if any VIOSes
                       are vNIC failover-capable, non-failover-capable VIOSes
                       are excluded.
    :return: The filtered list of VIOS wrappers.
    :raise NotEnoughActiveVioses: If no active (including RMC) VIOSes can be
                                  found.
    :raise NoVNICCapableVIOSes: If none of the vioses are vNIC capable.
    :raise VNICFailoverNotSupportedVIOS: If redundancy > 1 and none of the
                                         vioses is vNIC failover capable.
    """
    # This raises if none are found
    vioses = tpar.get_active_vioses(adap, xag=[], vios_wraps=vioses,
                                    find_min=1)
    # Filter by vNIC capability
    vioses = [vios for vios in vioses if vios.vnic_capable]
    if not vioses:
        raise ex.NoVNICCapableVIOSes()

    # Filter by failover capability, if needed.
    # If any are failover-capable, use just those, regardless of redundancy.
    failover_only = [vios for vios in vioses if vios.vnic_failover_capable]
    if redundancy > 1 or any(failover_only):
        vioses = failover_only

    # At this point, if the list is empty, it's because no failover capability.
    if not vioses:
        raise ex.VNICFailoverNotSupportedVIOS(red=redundancy)

    return vioses


def _get_good_sriovs(sriov_adaps):
    """(Retrieve and) filter SR-IOV adapters to those Running in Sriov mode.

    :param sriov_adaps: List of SRIOVAdapter wrappers to filter by mode/state.
    :return: List of SR-IOV adapters in Running state and in Sriov mode.
    :raise NoRunningSharedSriovAdapters: If no SR-IOV adapters can be found in
                                         Sriov mode and Running state.
    """
    # Filter SRIOV adapters to those in the correct mode/state
    good_adaps = [sriov for sriov in sriov_adaps if
                  sriov.mode == card.SRIOVAdapterMode.SRIOV and
                  sriov.state == card.SRIOVAdapterState.RUNNING]
    if not good_adaps:
        raise ex.NoRunningSharedSriovAdapters(
            sriov_loc_mode_state='\n'.join([' | '.join([
                sriov.phys_loc_code, sriov.mode,
                sriov.state or '-']) for sriov in sriov_adaps]))

    LOG.debug('Found running/shared SR-IOV adapter(s): %s',
              str([sriov.phys_loc_code for sriov in good_adaps]))

    return good_adaps


def _get_good_pport_list(sriov_adaps, pports, capacity, redundancy,
                         check_link_status):
    """Get a list of SRIOV*PPort filtered by capacity and specified pports.

    Builds a list of pypowervm.wrappers.iocard.SRIOV*PPort from sriov_adaps
    such that:
    - Only ports whose location codes are listed in the pports param are
      considered.
    - Only ports with sufficient remaining capacity (per the capacity param, if
      specified; otherwise the port's min_granularity) are considered.

    :param sriov_adaps: A list of SRIOVAdapter wrappers whose mode is Sriov and
                        whose state is Running.
    :param pports: A list of string physical location codes of the physical
                   ports to consider.
    :param capacity: (float) Minimum capacity which must be available on each
                     backing device.  Must be between 0.0 and 1.0, and must be
                     a multiple of the min_granularity of *all* of the pports.
                     If None, available port capacity is validated using each
                     port's min_granularity.
    :param redundancy: The desired redundancy level (number of ports to
                       return).required.  If the filtered list has fewer than
                       this number of ports, InsufficientSRIOVCapacity is
                       raised.
    :param check_link_status: If True, ports with link-down status will not be
                              returned.  If False, link status is not checked.
    :raise InsufficientSRIOVCapacity: If the final list contains fewer than
                                      'redundancy' ports.
    :return: A filtered list of SRIOV*PPort wrappers.
    """
    def port_ok(port):
        pok = True
        # Is it in the candidate list?
        if port.loc_code not in pports:
            pok = False
        # Is the link state up
        if check_link_status and not port.link_status:
            pok = False
        # Does it have available logical ports?
        if port.cfg_lps >= port.cfg_max_lps:
            pok = False
        # Does it have capacity?
        des_cap = port.min_granularity
        if capacity is not None:
            # Must be at least min_granularity.
            des_cap = max(des_cap, capacity)
        if port.allocated_capacity + des_cap > 1.0:
            pok = False
        return pok

    pport_wraps = []
    for sriov in sriov_adaps:
        for pport in sriov.phys_ports:
            if port_ok(pport):
                pp2add = copy.deepcopy(pport)
                pport_wraps.append(pp2add)

    if len(pport_wraps) < redundancy:
        raise ex.InsufficientSRIOVCapacity(red=redundancy,
                                           found_vfs=len(pport_wraps))
    LOG.debug('Filtered list of physical ports: %s' %
              str([pport.loc_code for pport in pport_wraps]))
    return pport_wraps


def get_lpar_vnics(adapter):
    """Return a dict mapping LPAR wrappers to their VNIC feeds.

    :param adapter: The pypowervm.adapter.Adapter for REST API communication.
    :return: A dict of the form { LPAR: [VNIC, ...] }, where the keys are
             pypowervm.wrappers.logical_partition.LPAR and the values are lists
             of the pypowervm.wrappers.iocard.VNIC they own.
    """
    return {lpar: card.VNIC.get(adapter, parent=lpar) for lpar in
            tpar.get_partitions(adapter, lpars=True, vioses=False)}


def _vnics_using_pport(pport, lpar2vnics):
    """Determine (and warn about) usage of SRIOV physical port by VNICs.

    Ascertain whether an SRIOV physical port is being used as a backing device
    for any VNICs.  The method returns a list of warning messages for each such
    usage found.

    :param pport: pypowervm.wrappers.iocard.SRIOV*PPort wrapper to check.
    :param lpar2vnics: Dict of {LPAR: [VNIC, ...]} gleaned from get_lpar_vnics
    :return: A list of warning messages for found usages of the physical port.
             If no usages were found, the empty list is returned.
    """
    warnings = []
    for lpar, vnics in six.iteritems(lpar2vnics):
        for vnic in vnics:
            if any([backdev for backdev in vnic.back_devs if
                    backdev.sriov_adap_id == pport.sriov_adap_id and
                    backdev.pport_id == pport.port_id]):
                warnings.append(
                    _("SR-IOV Physical Port at location %(loc_code)s is "
                      "backing a vNIC belonging to LPAR %(lpar_name)s (LPAR "
                      "UUID: %(lpar_uuid)s; vNIC UUID: %(vnic_uuid)s).") %
                    {'loc_code': pport.loc_code, 'lpar_name': lpar.name,
                     'lpar_uuid': lpar.uuid, 'vnic_uuid': vnic.uuid})
    return warnings


def _vet_port_usage(sys_w, label_index):
    """Look for relabeled ports which are in use by vNICs.

    :param sys_w: pypowervm.wrappers.managed_system.System wrapper for the
                  host.
    :param label_index: Dict of { port_loc_code: port_label_before } mapping
                        the physical location code of each physical port to the
                        value of its label before changes were made.
    :return: A list of translated messages warning of relabeled ports which are
             in use by vNICs.
    """
    warnings = []
    lpar2vnics = None
    for sriovadap in sys_w.asio_config.sriov_adapters:
        for pport in sriovadap.phys_ports:
            # If the port is unused, it's fine
            if pport.cfg_lps == 0:
                continue
            # If the original port label was unset, no harm setting it.
            if not label_index[pport.loc_code]:
                continue
            # If the port label is unchanged, it's fine
            if pport.label == label_index[pport.loc_code]:
                continue
            # Now we have to check all the VNICs on all the LPARs.  Lazy-load
            # this, because it's expensive.
            if lpar2vnics is None:
                lpar2vnics = get_lpar_vnics(sys_w.adapter)
            warnings += _vnics_using_pport(pport, lpar2vnics)
    return warnings


@tx.entry_transaction
def safe_update_pports(sys_w, callback_func, force=False):
    """Retrying entry transaction for safe updates to SR-IOV physical ports.

    Usage:
        def changes(sys_w):
            for sriov in sys_w.asio_config.sriov_adapters:
                ...
                sriov.phys_ports[n].pport.label = some_new_label
                ...
                update_needed = True
                ...
            return update_needed

        sys_w = safe_update_pports(System.getter(adap), changes, force=maybe)

    The consumer passes a callback method which makes changes to the labels of
    the physical ports of the ManagedSystem's SR-IOV adapters.

    If the callback returns a False value (indicating that no update is
    necessary), safe_update_pports immediately returns the sys_w.

    If the callback returns a True value, safe_update_pports first checks
    whether any of the changed ports are in use by vNICs (see "Why vNICs?"
    below).

    If the force option is not True, and any uses were found, this method
    raises an exception whose text includes details about the found usages.
    Otherwise, the found usages are logged as warnings.

    Assuming no exception is raised, safe_update_pports attempts to update the
    sys_w wrapper with the REST server.  (The caller does *not* do the update.)
    If an etag mismatch is encountered, safe_update_pports refreshes the sys_w
    wrapper and retries, according to the semantics of entry_transaction.

    Why vNICs?
    Care must be taken when changing port labels on the fly because those
    labels are used by LPM to ensure that the LPAR on the target system gets
    equivalent connectivity.  Direct-attached VFs - either those belonging to
    VIOSes (e.g. for SEA) or to LPARs - mean the partition is not migratable,
    so the labels can be changed with impunity.  And the only way a VF is
    migratable is if it belongs to a vNIC on a migratable LPAR.

    :param sys_w: pypowervm.wrappers.managed_system.System wrapper or getter
                  thereof.
    :param callback_func: Method executing the actual changes on the sys_w.
                          The method must accept sys_w (a System wrapper) as
                          its only argument.  Its return value will be
                          interpreted as a boolean to determine whether to
                          perform the update() (True) or not (False).
    :param force: If False (the default) and any of the updated physical ports
                  are found to be in use by vNICs, the method will raise.  If
                  True, warnings are logged for each such usage, but the method
                  will succeed.
    :return: The (possibly-updated) sys_w.
    :raise CantUpdatePPortsInUse: If any of the relabeled physical ports are in
                                  use by vNICs *and* the force option is False.
    """
    with PPORT_MOD_LOCK.write_lock():
        # Build an index of port:label for comparison after setting
        label_index = {pport.loc_code: pport.label for sriovadap in
                       sys_w.asio_config.sriov_adapters for pport in
                       sriovadap.phys_ports}

        # Let caller make the pport changes.
        if not callback_func(sys_w):
            # No update needed.
            # sys_w may be what was passed in, or the result of the getter.
            return sys_w

        # If return is True, caller wants us to update().  For each port that
        # changed, check its usage
        warnings = _vet_port_usage(sys_w, label_index)
        if warnings and not force:
            raise ex.CantUpdatePPortsInUse(warnings=warnings)
        # We're going to do the update.  Log any found usages.
        if warnings:
            LOG.warning(_("Making changes to the following SR-IOV physical "
                          "port labels even though they are in use by vNICs:"))
            for warning in warnings:
                LOG.warning(warning)
        return sys_w.update()


def find_pports_for_portlabel(portlabel, adapter, msys=None):
    """Find SR-IOV physical ports based on the port label.

    :param portlabel: portlabel of the SR-IOV physical ports to find.
    :param adapter: The pypowervm adapter API interface.
    :param msys: pypowervm.wrappers.managed_system.System wrapper.If not
                 specified, it will be retrieved from the server.
    :return: List of SRIOVEthPPort or SRIOVConvPPort wrappers for the specified
             port label, or the empty list if no such port exists.
    """
    # Physical ports for the given physical network
    if msys is None:
        msys = ms.System.get(adapter)[0]
    pports = []
    for sriov in msys.asio_config.sriov_adapters:
        for pport_w in sriov.phys_ports:
            if (pport_w.label or 'default') == portlabel:
                pports.append(pport_w)
    return pports


def find_pport(sys_w, physloc):
    """Find an SR-IOV physical port based on its location code.

    :param sys_w: pypowervm.wrappers.managed_system.System wrapper of the host.
    :param physloc: Physical location code string (per SRIOV*PPort.loc_code) of
                    the SR-IOV physical port to find.
    :return: SRIOVEthPPort or SRIOVConvPPort wrapper with the specified
             location code, or None if no such port exists in sys_w.
    """
    for sriov in sys_w.asio_config.sriov_adapters:
        for pport in sriov.phys_ports:
            if pport.loc_code == physloc:
                return pport
    return None



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import fixtures
import importlib
import mock
import six

from pypowervm import traits as trt

# An anchor for traits we construct artificially so the session isn't
# garbage collected.
_mk_traits_sessions = []


def _mk_traits(local, hmc):
    """Mock a single APITraits configuration.

    :param local: Should the APITraits pretend to be local?  True or False.
    :param hmc: Should the APITraits pretend to be running against HMC (True)
                or PVM (False)?
    :return: APITraits instance with the specified behavior.
    """
    _sess = mock.Mock()
    _sess.use_file_auth = local
    _sess.mc_type = 'HMC' if hmc else 'PVM'

    # Traits use a weak ref to the session to avoid a circular reference
    # so anchor the mock session globally otherwise it'll be gone
    # right after we return it.
    global _mk_traits_sessions
    _mk_traits_sessions.append(_sess)
    return trt.APITraits(_sess)

LocalPVMTraits = _mk_traits(local=True, hmc=False)
RemotePVMTraits = _mk_traits(local=False, hmc=False)
RemoteHMCTraits = _mk_traits(local=False, hmc=True)


class SessionFx(fixtures.Fixture):
    """Patch pypowervm.adapter.Session."""

    def __init__(self, traits=LocalPVMTraits):
        """Create Session patcher with traits.

        :param traits: APITraits instance to be assigned to the .traits
                       attribute of the mock Session.  If not specified,
                       sess.traits will be LocalPVMTraits. LocalPVMTraits,
                       RemotePVMTraits, and RemoteHMCTraits are provided below
                       for convenience.
        :return:
        """
        self.traits = traits
        self._patcher = mock.patch('pypowervm.adapter.Session')

    def setUp(self):
        super(SessionFx, self).setUp()
        self.sess = self._patcher.start()
        self.addCleanup(self._patcher.stop)
        self.sess.traits = self.traits
        self.sess.timeout = 1200


class AdapterFx(fixtures.Fixture):
    """Patch pypowervm.adapter.Adapter."""

    def __init__(self, session=None, traits=None):
        """Create Adapter and/or Session patchers with traits.

        :param session: A pypowervm.adapter.Session instance or mock with which
                        to back this mocked Adapter.  If not specified, a new
                        SessionFx fixture is created and used.
        :param traits: APITraits instance to be assigned to the .traits
                       attribute of the Session and/or Adapter mock.  If
                       not specified, LocalPVMTraits will be used.  If both
                       session and traits are specified, the session's traits
                       will be overwritten with the traits parameter.
                       LocalPVMTraits, RemotePVMTraits, and RemoteHMCTraits are
                       provided below for convenience.
        """
        super(AdapterFx, self).__init__()
        self.session = session
        if traits is None and (session is None or session.traits is None):
            self.traits = LocalPVMTraits
        elif traits:
            self.traits = traits
        else:
            self.traits = session.traits
        self._patcher = mock.patch('pypowervm.adapter.Adapter')

    def setUp(self):
        super(AdapterFx, self).setUp()
        if not self.session:
            self.session = self.useFixture(SessionFx(self.traits)).sess

        self.adpt = self._patcher.start()
        self.addCleanup(self._patcher.stop)
        self.adpt.session = self.session
        self.set_traits(self.traits)

    def set_traits(self, traits):
        # Mocked Adapter needs to see both routes to traits.
        self.adpt.session.traits = traits
        self.adpt.traits = traits


class SimplePatcher(object):
    """Provide a basic mocking patcher on a test fixture.

    The main purpose of this class is to be used with SimplePatchingFx.

    That said, the following are equivalent:

    @mock.patch('path.to.method')
    def test_foo(self, mock_meth):
        mock_meth.return_value = 'abc'
        # ...

    def test_foo(self):
        mock_meth = SimplePatcher(self, 'whatever', 'path.to.method',
                                  return_value='abc').start()
        # ...
    """
    def __init__(self, fx, name, path, patch_object=False, side_effect=None,
                 return_value=None):
        """Create a patcher on a given fixture.

        :param fx: The fixtures.Fixture (subclass) on which to register the
                   patcher.
        :param name: String name for the patcher.
        :param path: String python path of the object being mocked.
        :param patch_object: If True, the path parameter is parsed to create a
                             mock.patch.object with autospec=True instead of a
                             regular mock.patch.  For example,
                         patch='foo.bar.Baz.meth'
                             would result in
                         mock.patch.object(foo.bar.Baz, 'meth', autospec=True)
                             Note that this means the mock call will include
                             the instance through which it was invoked.
        :param side_effect: Side effect for the mock created by this patcher.
                            If side_effect is supplied, return_value is
                            ignored.
        :param return_value: Return value for the mock created by this patcher.
                             If side_effect is supplied, return_value is
                             ignored.
        """
        self.fx = fx
        self.name = name
        if patch_object:
            modname, klassname, methname = path.rsplit('.', 2)
            module = importlib.import_module(modname)
            klass = getattr(module, klassname)
            self.patcher = mock.patch.object(klass, methname, autospec=True)
        else:
            self.patcher = mock.patch(path)
        self.return_value = return_value
        self.side_effect = side_effect
        self.mock = None

    def start(self):
        """Start the patcher, creating the and setting up the mock."""
        self.mock = self.patcher.start()
        if self.side_effect:
            self.mock.side_effect = self.side_effect
        else:
            self.mock.return_value = self.return_value
        self.fx.addCleanup(self.patcher.stop)
        return self.mock


class LoggingPatcher(SimplePatcher):
    """SimplePatcher whose mock logs its name and returns a value."""
    FIRST_ARG = '__MOCK_RETURNS_FIRST_ARGUMENT__'

    def __init__(self, fx, name, path, patch_object=False, return_value=None):
        """Create the logging patcher.

        :param fx: The fixtures.Fixture (subclass) on which to register the
                   patcher.  Must be a fixture providing a .log(msg) method.
        :param name: String name for the patcher.
        :param path: String python path of the object being mocked.
        :param patch_object: If True, the path parameter is parsed to create a
                             mock.patch.object with autospec=True instead of a
                             regular mock.patch.  For example,
                         patch='foo.bar.Baz.meth'
                             would result in
                         mock.patch.object(foo.bar.Baz, 'meth', autospec=True)
                             Note that this means the mock call will include
                             the instance through which it was invoked.
        :param return_value: The return value for the mocked method.
        """
        def _log(*a, **k):
            self.fx.log(self.name)
            return a[0] if self.ret is self.FIRST_ARG else self.ret
        # This ignores/overrides the superclass's return_value semantic.
        self.ret = return_value
        super(LoggingPatcher, self).__init__(
            fx, name, path, patch_object=patch_object, side_effect=_log)


@six.add_metaclass(abc.ABCMeta)
class Logger(object):
    """Base class for mixins wanting simple 'log to a list' semantics."""
    def __init__(self):
        """Create a new Logger."""
        super(Logger, self).__init__()
        self._tx_log = []

    def get_log(self):
        """Retrieve the event log.

        :return: The log, a list of strings in the order they were added.
        """
        return self._tx_log

    def log(self, val):
        """Add a message to the log.

        :param val: String value to append to the log.
        """
        self._tx_log.append(val)

    def reset_log(self):
        """Clear the log."""
        self._tx_log = []


@six.add_metaclass(abc.ABCMeta)
class SimplePatchingFx(fixtures.Fixture):
    """Fixture base class supporting SimplePatcher.

    Subclasses should invoke add_patchers from __init__ after super().__init__,
    but before useFixture.
    """
    def __init__(self):
        """Create the simple-patching fixture."""
        super(SimplePatchingFx, self).__init__()
        self.patchers = {}

    def add_patchers(self, *patchers):
        """Add some number of SimplePatcher instances to the fixture.

        :param patchers: Zero or more SimplePatcher instances to add.
        """
        for patcher in patchers:
            self.patchers[patcher.name] = patcher

    def setUp(self):
        """Start the fixture and its member SimplePatchers.

        This is generally invoked via useFixture and should not be called
        directly.
        """
        super(SimplePatchingFx, self).setUp()
        for patcher in self.patchers.values():
            patcher.start()


class SleepPatcher(SimplePatcher):
    def __init__(self, fx, side_effect=None):
        super(SleepPatcher, self).__init__(fx, 'sleep', 'time.sleep',
                                           side_effect=side_effect)


class SleepFx(SimplePatchingFx):
    """Fixture for time.sleep."""
    def __init__(self, side_effect=None):
        """Create the fixture for time.sleep."""
        super(SleepFx, self).__init__()
        self.add_patchers(SleepPatcher(self, side_effect=side_effect))

# Thread locking primitives are located slightly differently in py2 vs py3
SEM_ENTER = 'threading.%sSemaphore.__enter__' % ('_' if six.PY2 else '')
SEM_EXIT = 'threading.%sSemaphore.__exit__' % ('_' if six.PY2 else '')


class WrapperTaskFx(SimplePatchingFx, Logger):
    """Customizable mocking and pseudo-logging for WrapperTask primitives.

    Provides LoggingPatchers for REST and locking primitives.  By default,
    these patchers simply log their name and return a sensible value (see
    below).

    However, patchers can be added, changed, or removed by name from the
    fixture instance via its 'patchers' dict.  In order to have effect on your
    test case, such modifications must be done between fixture initialization
    and useFixture.  For example:

    # Init the fixture, but do not start it:
    wtfx = WrapperTaskFx(a_wrapper)
    # An existing patcher can be modified:
    upd = wtfx.patchers['update'].side_effect = SomeException()
    # Or deleted:
    del wtfx.patchers['refresh']
    # New patchers can be added.  They must be instances of SimplePatcher (or a
    # subclass).  Add directly to 'patchers':
    wtfx.patchers['foo'] = LoggingPatcher(wtfx, 'frob', 'pypowervm.utils.frob')
    # ...or use add_patchers to add more than one:
    wtfx.add_patchers(p1, p2, p3)

    # Finally, don't forget to start the fixture
    self.useFixture(wtfx)

    # Mocks can be accessed via their patchers and queried during testing as
    # usual:
    wtfx.patchers['foo'].mock.assert_called_with('bar', 'baz')
    self.assertEqual(3, wtfx.patchers['update'].mock.call_count)

    See live examples in pypowervm.tests.utils.test_transaction.TestWrapperTask

    Default mocks:
    'get': Mocks EntyrWrapperGetter.get.
           Logs 'get'.
           Returns the wrapper with which the fixture was initialized.
    'refresh': Mocks EntryWrapper.refresh.
               Logs 'refresh'.
               Returns the wrapper with which the fixture was initialized.
    'update': Mocks EntryWrapper.update.
              Logs 'update'.
              Returns the wrapper with which the fixture was initialized.
    'lock', 'unlock': Mocks semaphore locking (oslo_concurrency.lockutils.lock
                      and @synchronized, ultimately threading.Semaphore)
                      performed by the @entry_transaction decorator.
                      Logs 'lock'/'unlock', respectively.
                      Returns None.

    """
    def __init__(self, wrapper):
        """Create the fixture around a specific EntryWrapper.

        :param wrapper: EntryWrapper instance to be returned by mocked
                        EntryWrapperGetter.get and EntryWrapper.refresh methods
        """
        super(WrapperTaskFx, self).__init__()
        self._wrapper = wrapper
        self.add_patchers(
            LoggingPatcher(
                self, 'get',
                'pypowervm.wrappers.entry_wrapper.EntryWrapperGetter.get',
                return_value=self._wrapper),
            LoggingPatcher(
                self, 'refresh',
                'pypowervm.wrappers.entry_wrapper.EntryWrapper.refresh',
                return_value=self._wrapper),
            LoggingPatcher(
                self, 'update',
                'pypowervm.wrappers.entry_wrapper.EntryWrapper.update',
                return_value=self._wrapper),
            LoggingPatcher(self, 'lock', SEM_ENTER),
            LoggingPatcher(self, 'unlock', SEM_EXIT),
            SleepPatcher(self)
        )


class FeedTaskFx(SimplePatchingFx, Logger):
    """Customizable mocking and pseudo-logging for FeedTask primitives.

    Provides LoggingPatchers for REST and locking primitives.  By default,
    these patchers simply log their name and return a sensible value (see
    below).

    However, patchers can be added, changed, or removed by name from the
    fixture instance via its 'patchers' dict.  In order to have effect on your
    test case, such modifications must be done between fixture initialization
    and useFixture.  For example:

    # Init the fixture, but do not start it:
    ftfx = FeedTaskFx(a_feed)
    # An existing patcher can be modified:
    upd = ftfx.patchers['update'].side_effect = SomeException()
    # Or deleted:
    del ftfx.patchers['refresh']
    # New patchers can be added.  They must be instances of SimplePatcher (or a
    # subclass).  Add directly to 'patchers':
    ftfx.patchers['foo'] = LoggingPatcher(ftfx, 'frob', 'pypowervm.utils.frob')
    # ...or use add_patchers to add more than one:
    ftfx.add_patchers(p1, p2, p3)

    # Finally, don't forget to start the fixture
    self.useFixture(ftfx)

    # Mocks can be accessed via their patchers and queried during testing as
    # usual:
    ftfx.patchers['foo'].mock.assert_called_with('bar', 'baz')
    self.assertEqual(3, ftfx.patchers['update'].mock.call_count)

    See live examples in pypowervm.tests.utils.test_transaction.TestWrapperTask

    Default mocks:
    'get': Mocks FeedGetter.get.
           Logs 'get'.
           Returns the feed with which the fixture was initialized.
    'refresh': Mocks EntryWrapper.refresh.
               Logs 'refresh'.
               Returns the wrapper on which the refresh method was called.
    'update': Mocks EntryWrapper.update.
              Logs 'update'.
              Returns the wrapper on which the update method was called.
    'lock', 'unlock': Mocks semaphore locking (oslo_concurrency.lockutils.lock
                      and @synchronized, ultimately threading.Semaphore)
                      performed by the @entry_transaction decorator.
                      Logs 'lock'/'unlock', respectively.
                      Returns None.
    """
    def __init__(self, feed):
        """Create the fixture around a given feed.

        :param feed: The feed (list of EntryWrappers) to be returned from the
                     mocked FeedGetter.get method.
        """
        super(FeedTaskFx, self).__init__()
        self._feed = feed
        self.add_patchers(
            LoggingPatcher(
                self, 'get',
                'pypowervm.wrappers.entry_wrapper.FeedGetter.get',
                return_value=self._feed),
            LoggingPatcher(
                self, 'refresh',
                'pypowervm.wrappers.entry_wrapper.EntryWrapper.refresh',
                patch_object=True, return_value=LoggingPatcher.FIRST_ARG),
            LoggingPatcher(
                self, 'update',
                'pypowervm.wrappers.entry_wrapper.EntryWrapper.update',
                patch_object=True, return_value=LoggingPatcher.FIRST_ARG),
            LoggingPatcher(self, 'lock', SEM_ENTER),
            LoggingPatcher(self, 'unlock', SEM_EXIT),
            SleepPatcher(self)
        )


class LoggingFx(SimplePatchingFx):
    """Fixture for LOG.*, not to be confused with Logger/LoggingPatcher.

    Provides patches and mocks for LOG.x for x in
    ('info', 'warning', 'debug', 'error', 'exception')
    """
    def __init__(self):
        """Create the fixture for the various logging methods."""
        super(LoggingFx, self).__init__()
        self.add_patchers(
            *(SimplePatcher(self, x, 'oslo_log.log.BaseLoggerAdapter.%s' % x)
              for x in ('info', 'warning', 'debug', 'error', 'exception')))
